@@ -26,7 +26,7 @@
 #ifndef _COMPIZ_H
 #define _COMPIZ_H
 
-#define ABIVERSION 20061020
+#define ABIVERSION 20070213
 
 #include <stdio.h>
 #include <sys/time.h>
@@ -46,6 +46,10 @@
 #include <GL/gl.h>
 #include <GL/glx.h>
 
+#ifdef  __cplusplus
+extern "C" {
+#endif
+
 #if COMPOSITE_MAJOR > 0 || COMPOSITE_MINOR > 2
 #define USE_COW
 #endif
@@ -69,12 +73,17 @@
 #define  _(x) x
 #define N_(x) x
 
-typedef struct _CompPlugin  CompPlugin;
-typedef struct _CompDisplay CompDisplay;
-typedef struct _CompScreen  CompScreen;
-typedef struct _CompWindow  CompWindow;
-typedef struct _CompTexture CompTexture;
-typedef struct _CompIcon    CompIcon;
+typedef struct _CompPlugin	  CompPlugin;
+typedef struct _CompDisplay	  CompDisplay;
+typedef struct _CompScreen	  CompScreen;
+typedef struct _CompWindow	  CompWindow;
+typedef struct _CompTexture	  CompTexture;
+typedef struct _CompIcon	  CompIcon;
+typedef struct _CompWindowExtents CompWindowExtents;
+typedef struct _CompProgram	  CompProgram;
+typedef struct _CompFunction	  CompFunction;
+typedef struct _CompFunctionData  CompFunctionData;
+typedef struct _FragmentAttrib    FragmentAttrib;
 
 /* virtual modifiers */
 
@@ -137,15 +146,16 @@ typedef struct _CompIcon    CompIcon;
 #define MAXIMIZE_STATE (CompWindowStateMaximizedHorzMask | \
 			CompWindowStateMaximizedVertMask)
 
-#define CompWindowActionMoveMask	 (1 << 0)
-#define CompWindowActionResizeMask	 (1 << 1)
-#define CompWindowActionStickMask	 (1 << 2)
-#define CompWindowActionMinimizeMask     (1 << 3)
-#define CompWindowActionMaximizeHorzMask (1 << 4)
-#define CompWindowActionMaximizeVertMask (1 << 5)
-#define CompWindowActionFullscreenMask	 (1 << 6)
-#define CompWindowActionCloseMask	 (1 << 7)
-#define CompWindowActionShadeMask	 (1 << 8)
+#define CompWindowActionMoveMask	  (1 << 0)
+#define CompWindowActionResizeMask	  (1 << 1)
+#define CompWindowActionStickMask	  (1 << 2)
+#define CompWindowActionMinimizeMask      (1 << 3)
+#define CompWindowActionMaximizeHorzMask  (1 << 4)
+#define CompWindowActionMaximizeVertMask  (1 << 5)
+#define CompWindowActionFullscreenMask	  (1 << 6)
+#define CompWindowActionCloseMask	  (1 << 7)
+#define CompWindowActionShadeMask	  (1 << 8)
+#define CompWindowActionChangeDesktopMask (1 << 9)
 
 #define MwmFuncAll      (1L << 0)
 #define MwmFuncResize   (1L << 1)
@@ -203,6 +213,7 @@ extern Bool       replaceCurrentWm;
 extern Bool       indirectRendering;
 extern Bool       strictBinding;
 extern Bool       useCow;
+extern Bool       noDetection;
 
 extern int  defaultRefreshRate;
 extern char *defaultTextureFilter;
@@ -251,33 +262,6 @@ freePrivateIndex (int  len,
 		  int  index);
 
 
-/* readpng.c */
-
-Bool
-openImageFile (const char *filename,
-	       char	  **returnFilename,
-	       FILE	  **returnFile);
-
-Bool
-readPng (const char   *filename,
-	 char	      **data,
-	 unsigned int *width,
-	 unsigned int *height);
-
-Bool
-readPngBuffer (const unsigned char *buffer,
-	       char		   **data,
-	       unsigned int	   *width,
-	       unsigned int	   *height);
-
-Bool
-writePngToFile (unsigned char *buffer,
-		char	      *filename,
-		int	      width,
-		int	      height,
-		int	      stride);
-
-
 /* option.c */
 
 typedef enum {
@@ -291,9 +275,10 @@ typedef enum {
 } CompOptionType;
 
 typedef enum {
-    CompBindingTypeNone   = 0,
-    CompBindingTypeKey    = 1 << 0,
-    CompBindingTypeButton = 1 << 1
+    CompBindingTypeNone       = 0,
+    CompBindingTypeKey        = 1 << 0,
+    CompBindingTypeButton     = 1 << 1,
+    CompBindingTypeEdgeButton = 1 << 2
 } CompBindingType;
 
 typedef enum {
@@ -344,6 +329,7 @@ struct _CompAction {
     Bool bell;
 
     unsigned int edgeMask;
+    int		 edgeButton;
 };
 
 typedef struct {
@@ -457,6 +443,12 @@ getStringOptionNamed (CompOption *option,
 		      char	 *name,
 		      char	 *defaultValue);
 
+unsigned short *
+getColorOptionNamed (CompOption	    *option,
+		     int	    nOption,
+		     char	    *name,
+		     unsigned short *defaultValue);
+
 char *
 keyBindingToString (CompDisplay    *d,
 		    CompKeyBinding *key);
@@ -485,11 +477,15 @@ stringToColor (const char     *color,
 char *
 colorToString (unsigned short *rgba);
 
+char *
+optionTypeToString (CompOptionType type);
+
 
 /* display.c */
 
 typedef int CompTimeoutHandle;
 typedef int CompWatchFdHandle;
+typedef int CompFileWatchHandle;
 
 #define COMP_DISPLAY_OPTION_ACTIVE_PLUGINS                0
 #define COMP_DISPLAY_OPTION_TEXTURE_FILTER                1
@@ -547,7 +543,10 @@ typedef int CompWatchFdHandle;
 #define COMP_DISPLAY_OPTION_HIDE_SKIP_TASKBAR_WINDOWS     53
 #define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_SHADED          54
 #define COMP_DISPLAY_OPTION_IGNORE_HINTS_WHEN_MAXIMIZED   55
-#define COMP_DISPLAY_OPTION_NUM                           56
+#define COMP_DISPLAY_OPTION_TERMINAL			  56
+#define COMP_DISPLAY_OPTION_RUN_TERMINAL		  57
+#define COMP_DISPLAY_OPTION_PING_DELAY			  58
+#define COMP_DISPLAY_OPTION_NUM				  59
 
 typedef CompOption *(*GetDisplayOptionsProc) (CompDisplay *display,
 					      int	  *count);
@@ -568,11 +567,56 @@ typedef void (*FiniPluginForDisplayProc) (CompPlugin  *plugin,
 typedef void (*HandleEventProc) (CompDisplay *display,
 				 XEvent	     *event);
 
+typedef void (*HandleCompizEventProc) (CompDisplay *display,
+				       char        *pluginName,
+				       char        *eventName,
+				       CompOption  *option,
+				       int         nOption);
+
 typedef Bool (*CallBackProc) (void *closure);
 
 typedef void (*ForEachWindowProc) (CompWindow *window,
 				   void	      *closure);
 
+typedef Bool (*FileToImageProc) (CompDisplay *display,
+				 const char  *path,
+				 const char  *name,
+				 int	     *width,
+				 int	     *height,
+				 int	     *stride,
+				 void	     **data);
+
+typedef Bool (*ImageToFileProc) (CompDisplay *display,
+				 const char  *path,
+				 const char  *name,
+				 const char  *format,
+				 int	     width,
+				 int	     height,
+				 int	     stride,
+				 void	     *data);
+
+#define NOTIFY_CREATE_MASK (1 << 0)
+#define NOTIFY_DELETE_MASK (1 << 1)
+#define NOTIFY_MOVE_MASK   (1 << 2)
+
+typedef void (*FileWatchCallBackProc) (const char *name,
+				       void	  *closure);
+
+typedef struct _CompFileWatch {
+    struct _CompFileWatch *next;
+    char		  *path;
+    int			  mask;
+    FileWatchCallBackProc callBack;
+    void		  *closure;
+    CompFileWatchHandle   handle;
+} CompFileWatch;
+
+typedef void (*FileWatchAddedProc) (CompDisplay	  *display,
+				    CompFileWatch *fileWatch);
+
+typedef void (*FileWatchRemovedProc) (CompDisplay   *display,
+				      CompFileWatch *fileWatch);
+
 struct _CompDisplay {
     Display    *display;
     CompScreen *screens;
@@ -659,6 +703,7 @@ struct _CompDisplay {
     Atom winActionFullscreenAtom;
     Atom winActionCloseAtom;
     Atom winActionShadeAtom;
+    Atom winActionChangeDesktopAtom;
 
     Atom wmAllowedActionsAtom;
 
@@ -708,6 +753,8 @@ struct _CompDisplay {
     Atom xdndEnterAtom;
     Atom xdndLeaveAtom;
     Atom xdndPositionAtom;
+    Atom xdndStatusAtom;
+    Atom xdndDropAtom;
 
     Atom managerAtom;
     Atom targetsAtom;
@@ -743,13 +790,22 @@ struct _CompDisplay {
     CompOptionValue plugin;
     Bool	    dirtyPluginList;
 
+    CompFileWatch *fileWatch;
+
     SetDisplayOptionProc	  setDisplayOption;
     SetDisplayOptionForPluginProc setDisplayOptionForPlugin;
 
     InitPluginForDisplayProc initPluginForDisplay;
     FiniPluginForDisplayProc finiPluginForDisplay;
 
-    HandleEventProc handleEvent;
+    HandleEventProc       handleEvent;
+    HandleCompizEventProc handleCompizEvent;
+
+    FileToImageProc fileToImage;
+    ImageToFileProc imageToFile;
+
+    FileWatchAddedProc   fileWatchAdded;
+    FileWatchRemovedProc fileWatchRemoved;
 
     CompPrivate *privates;
 };
@@ -783,6 +839,25 @@ compAddWatchFd (int	     fd,
 void
 compRemoveWatchFd (CompWatchFdHandle handle);
 
+CompFileWatchHandle
+addFileWatch (CompDisplay	    *display,
+	      const char	    *path,
+	      int		    mask,
+	      FileWatchCallBackProc callBack,
+	      void		    *closure);
+
+void
+removeFileWatch (CompDisplay	     *display,
+		 CompFileWatchHandle handle);
+
+void
+fileWatchAdded (CompDisplay   *display,
+		CompFileWatch *fileWatch);
+
+void
+fileWatchRemoved (CompDisplay   *display,
+		  CompFileWatch *fileWatch);
+
 int
 compCheckForError (Display *dpy);
 
@@ -850,6 +925,41 @@ setDisplayAction (CompDisplay     *display,
 		  CompOption      *o,
 		  CompOptionValue *value);
 
+Bool
+readImageFromFile (CompDisplay *display,
+		   const char  *name,
+		   int	       *width,
+		   int	       *height,
+		   void	       **data);
+
+Bool
+writeImageToFile (CompDisplay *display,
+		  const char  *path,
+		  const char  *name,
+		  const char  *format,
+		  int	      width,
+		  int	      height,
+		  void	      *data);
+
+Bool
+fileToImage (CompDisplay *display,
+	     const char	 *path,
+	     const char	 *name,
+	     int	 *width,
+	     int	 *height,
+	     int	 *stride,
+	     void	 **data);
+
+Bool
+imageToFile (CompDisplay *display,
+	     const char	 *path,
+	     const char	 *name,
+	     const char	 *format,
+	     int	 width,
+	     int	 height,
+	     int	 stride,
+	     void	 *data);
+
 
 /* event.c */
 
@@ -857,6 +967,13 @@ void
 handleEvent (CompDisplay *display,
 	     XEvent      *event);
 
+void
+handleCompizEvent (CompDisplay *display,
+		   char        *pluginName,
+		   char        *eventName,
+		   CompOption  *option,
+		   int         nOption);
+
 void
 handleSyncAlarm (CompWindow *w);
 
@@ -870,6 +987,10 @@ eventTerminates (CompDisplay *display,
 		 XEvent      *event,
 		 CompOption  *option);
 
+void
+clearTargetOutput (CompDisplay	*display,
+		   unsigned int mask);
+
 /* paint.c */
 
 #define MULTIPLY_USHORT(us1, us2)		 \
@@ -878,6 +999,11 @@ eventTerminates (CompDisplay *display,
 /* camera distance from screen, 0.5 * tan (FOV) */
 #define DEFAULT_Z_CAMERA 0.866025404f
 
+typedef struct _CompTransform {
+    float m[16];
+} CompTransform;
+
+/* XXX: ScreenPaintAttrib will be removed */
 typedef struct _ScreenPaintAttrib {
     GLfloat xRotate;
     GLfloat yRotate;
@@ -888,12 +1014,15 @@ typedef struct _ScreenPaintAttrib {
     GLfloat zCamera;
 } ScreenPaintAttrib;
 
+/* XXX: scale and translate fields will be removed */
 typedef struct _WindowPaintAttrib {
     GLushort opacity;
     GLushort brightness;
     GLushort saturation;
     GLfloat  xScale;
     GLfloat  yScale;
+    GLfloat  xTranslate;
+    GLfloat  yTranslate;
 } WindowPaintAttrib;
 
 extern ScreenPaintAttrib defaultScreenPaintAttrib;
@@ -927,47 +1056,106 @@ typedef void (*DonePaintScreenProc) (CompScreen *screen);
 
 typedef Bool (*PaintScreenProc) (CompScreen		 *screen,
 				 const ScreenPaintAttrib *sAttrib,
+				 const CompTransform	 *transform,
 				 Region			 region,
 				 int		         output,
 				 unsigned int		 mask);
 
 typedef void (*PaintTransformedScreenProc) (CompScreen		    *screen,
 					    const ScreenPaintAttrib *sAttrib,
+					    const CompTransform	    *transform,
+					    Region		    region,
 					    int			    output,
 					    unsigned int	    mask);
 
+/* XXX: ApplyScreenTransformProc will be removed */
+typedef void (*ApplyScreenTransformProc) (CompScreen		  *screen,
+					  const ScreenPaintAttrib *sAttrib,
+					  int			  output,
+					  CompTransform	          *transform);
+
+/*
+  window paint flags
+
+  bit 1-16 are used for read-only flags and they provide
+  information that describe the screen rendering pass
+  currently in process.
+
+  bit 17-32 are writable flags and they provide information
+  that is used to optimize rendering.
+*/
+
+/*
+  this flag is present when we are painting clipped
+  opaque window regions only.
+*/
+#define PAINT_WINDOW_CLIP_OPAQUE_MASK		(1 << 0)
+
+/*
+  this flag is present when we are painting clipped
+  translucent window regions only.
+*/
+#define PAINT_WINDOW_CLIP_TRANSLUCENT_MASK	(1 << 1)
+
+/*
+  this flag is present when window is being painted
+  on a transformed screen.
+*/
+#define PAINT_WINDOW_ON_TRANSFORMED_SCREEN_MASK (1 << 2)
+
+/*
+  flag indicate that window is translucent.
+*/
+#define PAINT_WINDOW_TRANSLUCENT_MASK           (1 << 16)
+
+/*
+  flag indicate that window is transformed.
+*/
+#define PAINT_WINDOW_TRANSFORMED_MASK           (1 << 17)
+
+/*
+  flag indicate that core PaintWindow function should
+  not draw this window.
+*/
+#define PAINT_WINDOW_NO_CORE_INSTANCE_MASK	(1 << 18)
+
+/*
+  flag indicate that blending is required.
+*/
+#define PAINT_WINDOW_BLEND_MASK			(1 << 19)
 
-#define PAINT_WINDOW_SOLID_MASK			(1 << 0)
-#define PAINT_WINDOW_TRANSLUCENT_MASK		(1 << 1)
-#define PAINT_WINDOW_TRANSFORMED_MASK           (1 << 2)
-#define PAINT_WINDOW_ON_TRANSFORMED_SCREEN_MASK (1 << 3)
 
 typedef Bool (*PaintWindowProc) (CompWindow		 *window,
 				 const WindowPaintAttrib *attrib,
+				 const CompTransform     *transform,
 				 Region			 region,
 				 unsigned int		 mask);
 
+typedef Bool (*DrawWindowProc) (CompWindow	     *window,
+				const CompTransform  *transform,
+				const FragmentAttrib *fragment,
+				Region		     region,
+				unsigned int	     mask);
+
 typedef void (*AddWindowGeometryProc) (CompWindow *window,
 				       CompMatrix *matrix,
 				       int	  nMatrix,
 				       Region	  region,
 				       Region	  clip);
 
-typedef void (*DrawWindowTextureProc) (CompWindow	       *w,
-				       CompTexture	       *texture,
-				       const WindowPaintAttrib *attrib,
-				       unsigned int	       mask);
+typedef void (*DrawWindowTextureProc) (CompWindow	    *w,
+				       CompTexture	    *texture,
+				       const FragmentAttrib *fragment,
+				       unsigned int	    mask);
 
 typedef void (*DrawWindowGeometryProc) (CompWindow *window);
 
 #define PAINT_BACKGROUND_ON_TRANSFORMED_SCREEN_MASK (1 << 0)
-#define PAINT_BACKGROUND_WITH_STENCIL_MASK          (1 << 1)
 
 typedef void (*PaintBackgroundProc) (CompScreen   *screen,
 				     Region	  region,
 				     unsigned int mask);
 
-
 void
 preparePaintScreen (CompScreen *screen,
 		    int	       msSinceLastPaint);
@@ -976,17 +1164,36 @@ void
 donePaintScreen (CompScreen *screen);
 
 void
-translateRotateScreen (const ScreenPaintAttrib *sa);
+transformToScreenSpace (CompScreen    *screen,
+			int	      output,
+			float         z,
+			CompTransform *transform);
+
+/* XXX: prepareXCoords will be removed */
+void
+prepareXCoords (CompScreen *screen,
+		int	   output,
+		float      z);
 
 void
 paintTransformedScreen (CompScreen		*screen,
 			const ScreenPaintAttrib *sAttrib,
+			const CompTransform	*transform,
+			Region			region,
 			int			output,
 			unsigned int	        mask);
 
+/* XXX: applyScreenTransform will be removed */
+void
+applyScreenTransform (CompScreen	      *screen,
+		      const ScreenPaintAttrib *sAttrib,
+		      int		      output,
+		      CompTransform	      *transform);
+
 Bool
 paintScreen (CompScreen		     *screen,
 	     const ScreenPaintAttrib *sAttrib,
+	     const CompTransform     *transform,
 	     Region		     region,
 	     int		     output,
 	     unsigned int	     mask);
@@ -1007,17 +1214,22 @@ addWindowGeometry (CompWindow *w,
 		   Region     clip);
 
 void
-drawWindowGeometry (CompWindow *w);
+drawWindowTexture (CompWindow		*w,
+		   CompTexture		*texture,
+		   const FragmentAttrib	*fragment,
+		   unsigned int		mask);
 
-void
-drawWindowTexture (CompWindow		   *w,
-		   CompTexture		   *texture,
-		   const WindowPaintAttrib *attrib,
-		   unsigned int		   mask);
+Bool
+drawWindow (CompWindow		 *w,
+	    const CompTransform  *transform,
+	    const FragmentAttrib *fragment,
+	    Region		 region,
+	    unsigned int	 mask);
 
 Bool
 paintWindow (CompWindow		     *w,
 	     const WindowPaintAttrib *attrib,
+	     const CompTransform     *transform,
 	     Region		     region,
 	     unsigned int	     mask);
 
@@ -1071,13 +1283,6 @@ readImageToTexture (CompScreen   *screen,
 		    unsigned int *width,
 		    unsigned int *height);
 
-Bool
-readImageBufferToTexture (CompScreen	      *screen,
-			  CompTexture	      *texture,
-			  const unsigned char *imageBuffer,
-			  unsigned int	      *returnWidth,
-			  unsigned int	      *returnHeight);
-
 Bool
 iconToTexture (CompScreen *screen,
 	       CompIcon   *icon);
@@ -1125,7 +1330,11 @@ disableTexture (CompScreen  *screen,
 #define COMP_SCREEN_OPTION_UNREDIRECT_FS       6
 #define COMP_SCREEN_OPTION_DEFAULT_ICON        7
 #define COMP_SCREEN_OPTION_SYNC_TO_VBLANK      8
-#define COMP_SCREEN_OPTION_NUM		       9
+#define COMP_SCREEN_OPTION_NUMBER_OF_DESKTOPS  9
+#define COMP_SCREEN_OPTION_DETECT_OUTPUTS      10
+#define COMP_SCREEN_OPTION_OUTPUTS	       11
+#define COMP_SCREEN_OPTION_FOCUS_PREVENTION    12
+#define COMP_SCREEN_OPTION_NUM		       13
 
 #ifndef GLX_EXT_texture_from_pixmap
 #define GLX_BIND_TO_TEXTURE_RGB_EXT        0x20D0
@@ -1204,12 +1413,12 @@ typedef void (*GLProgramStringProc) (GLenum	  target,
 				     GLenum	  format,
 				     GLsizei	  len,
 				     const GLvoid *string);
-typedef void (*GLProgramLocalParameter4fProc) (GLenum  target,
-					       GLuint  index,
-					       GLfloat x,
-					       GLfloat y,
-					       GLfloat z,
-					       GLfloat w);
+typedef void (*GLProgramParameter4fProc) (GLenum  target,
+					  GLuint  index,
+					  GLfloat x,
+					  GLfloat y,
+					  GLfloat z,
+					  GLfloat w);
 
 typedef void (*GLGenFramebuffersProc) (GLsizei n,
 				       GLuint  *framebuffers);
@@ -1250,9 +1459,8 @@ typedef Bool (*DamageWindowRectProc) (CompWindow *w,
 typedef Bool (*DamageWindowRegionProc) (CompWindow *w,
 					Region     region);
 
-typedef void (*SetWindowScaleProc) (CompWindow *w,
-				    float      xScale,
-				    float      yScale);
+typedef void (*GetOutputExtentsForWindowProc) (CompWindow	 *w,
+					       CompWindowExtents *output);
 
 typedef Bool (*FocusWindowProc) (CompWindow *window);
 
@@ -1278,6 +1486,8 @@ typedef void (*WindowUngrabNotifyProc) (CompWindow *window);
 
 typedef void (*WindowStateChangeNotifyProc) (CompWindow *window);
 
+typedef void (*OutputChangeNotifyProc) (CompScreen *screen);
+
 #define COMP_SCREEN_DAMAGE_PENDING_MASK (1 << 0)
 #define COMP_SCREEN_DAMAGE_REGION_MASK  (1 << 1)
 #define COMP_SCREEN_DAMAGE_ALL_MASK     (1 << 2)
@@ -1346,8 +1556,11 @@ struct _CompIcon {
 };
 
 typedef struct _CompOutput {
-    char   *name;
-    REGION region;
+    char       *name;
+    REGION     region;
+    int        width;
+    int        height;
+    XRectangle workArea;
 } CompOutput;
 
 struct _CompScreen {
@@ -1367,6 +1580,8 @@ struct _CompScreen {
     int		      y;
     int		      hsize;		/* Number of horizontal viewports */
     int		      vsize;		/* Number of vertical viewports */
+    unsigned int      nDesktop;
+    unsigned int      currentDesktop;
     REGION	      region;
     Region	      damage;
     unsigned long     damageMask;
@@ -1390,6 +1605,7 @@ struct _CompScreen {
     int		      sizeExpose;
     int		      nExpose;
     CompTexture       backgroundTexture;
+    Bool	      backgroundLoaded;
     unsigned int      pendingDestroys;
     int		      desktopWindowCount;
     unsigned int      mapNum;
@@ -1447,7 +1663,16 @@ struct _CompScreen {
     int		   timeLeft;
     Bool	   pendingCommands;
 
-    GLint stencilRef;
+    int lastFunctionId;
+
+    CompFunction *fragmentFunctions;
+    CompProgram  *fragmentPrograms;
+
+    int saturateFunction[2][64];
+
+    GLfloat projection[16];
+
+    Bool clearBuffers;
 
     Bool lighting;
     Bool slowAnimations;
@@ -1458,6 +1683,9 @@ struct _CompScreen {
 
     unsigned int showingDesktopMask;
 
+    unsigned long *desktopHintData;
+    int           desktopHintSize;
+
     GLXGetProcAddressProc    getProcAddress;
     GLXBindTexImageProc      bindTexImage;
     GLXReleaseTexImageProc   releaseTexImage;
@@ -1472,11 +1700,12 @@ struct _CompScreen {
     GLActiveTextureProc       activeTexture;
     GLClientActiveTextureProc clientActiveTexture;
 
-    GLGenProgramsProc		  genPrograms;
-    GLDeleteProgramsProc	  deletePrograms;
-    GLBindProgramProc		  bindProgram;
-    GLProgramStringProc		  programString;
-    GLProgramLocalParameter4fProc programLocalParameter4f;
+    GLGenProgramsProc	     genPrograms;
+    GLDeleteProgramsProc     deletePrograms;
+    GLBindProgramProc	     bindProgram;
+    GLProgramStringProc	     programString;
+    GLProgramParameter4fProc programEnvParameter4f;
+    GLProgramParameter4fProc programLocalParameter4f;
 
     GLGenFramebuffersProc        genFramebuffers;
     GLDeleteFramebuffersProc     deleteFramebuffers;
@@ -1495,18 +1724,19 @@ struct _CompScreen {
     InitPluginForScreenProc initPluginForScreen;
     FiniPluginForScreenProc finiPluginForScreen;
 
-    PreparePaintScreenProc       preparePaintScreen;
-    DonePaintScreenProc	         donePaintScreen;
-    PaintScreenProc	         paintScreen;
-    PaintTransformedScreenProc   paintTransformedScreen;
-    PaintBackgroundProc          paintBackground;
-    PaintWindowProc	         paintWindow;
-    AddWindowGeometryProc        addWindowGeometry;
-    DrawWindowGeometryProc       drawWindowGeometry;
-    DrawWindowTextureProc	 drawWindowTexture;
-    DamageWindowRectProc         damageWindowRect;
-    FocusWindowProc		 focusWindow;
-    SetWindowScaleProc		 setWindowScale;
+    PreparePaintScreenProc	  preparePaintScreen;
+    DonePaintScreenProc		  donePaintScreen;
+    PaintScreenProc		  paintScreen;
+    PaintTransformedScreenProc	  paintTransformedScreen;
+    ApplyScreenTransformProc	  applyScreenTransform;
+    PaintBackgroundProc		  paintBackground;
+    PaintWindowProc		  paintWindow;
+    DrawWindowProc		  drawWindow;
+    AddWindowGeometryProc	  addWindowGeometry;
+    DrawWindowTextureProc	  drawWindowTexture;
+    DamageWindowRectProc	  damageWindowRect;
+    GetOutputExtentsForWindowProc getOutputExtentsForWindow;
+    FocusWindowProc		  focusWindow;
 
     WindowResizeNotifyProc windowResizeNotify;
     WindowMoveNotifyProc   windowMoveNotify;
@@ -1515,6 +1745,8 @@ struct _CompScreen {
 
     WindowStateChangeNotifyProc windowStateChangeNotify;
 
+    OutputChangeNotifyProc outputChangeNotify;
+
     CompPrivate *privates;
 };
 
@@ -1544,6 +1776,12 @@ updateScreenBackground (CompScreen  *screen,
 void
 detectRefreshRateOfScreen (CompScreen *s);
 
+void
+showOutputWindow (CompScreen *s);
+
+void
+hideOutputWindow (CompScreen *s);
+
 Bool
 addScreen (CompDisplay *display,
 	   int	       screenNum,
@@ -1702,6 +1940,58 @@ outputDeviceForPoint (CompScreen *s,
 		      int	 x,
 		      int	 y);
 
+void
+getCurrentOutputExtents (CompScreen *s,
+			 int	    *x1,
+			 int	    *y1,
+			 int	    *x2,
+			 int	    *y2);
+
+void
+getWorkareaForOutput (CompScreen *s,
+		      int	 output,
+		      XRectangle *area);
+
+void
+setNumberOfDesktops (CompScreen   *s,
+		     unsigned int nDesktop);
+
+void
+setCurrentDesktop (CompScreen   *s,
+		   unsigned int desktop);
+
+void
+setDefaultViewport (CompScreen *s);
+
+void
+outputChangeNotify (CompScreen *s);
+
+void
+clearScreenOutput (CompScreen   *s,
+		   int	        output,
+		   unsigned int mask);
+
+void
+viewportForGeometry (CompScreen *s,
+		     int	x,
+		     int	y,
+		     int	width,
+		     int	height,
+		     int	borderWidth,
+		     int	*viewportX,
+		     int	*viewportY);
+
+int
+outputDeviceForGeometry (CompScreen *s,
+			 int	    x,
+			 int	    y,
+			 int	    width,
+			 int	    height,
+			 int	    borderWidth);
+
+Bool
+updateDefaultIcon (CompScreen *screen);
+
 
 /* window.c */
 
@@ -1718,12 +2008,12 @@ typedef Bool (*InitPluginForWindowProc) (CompPlugin *plugin,
 typedef void (*FiniPluginForWindowProc) (CompPlugin *plugin,
 					 CompWindow *window);
 
-typedef struct _CompWindowExtents {
+struct _CompWindowExtents {
     int left;
     int right;
     int top;
     int bottom;
-} CompWindowExtents;
+};
 
 typedef struct _CompStruts {
     XRectangle left;
@@ -1773,19 +2063,26 @@ struct _CompWindow {
     Bool	      damaged;
     Bool	      redirected;
     Bool	      managed;
+    Bool	      bindFailed;
     int		      destroyRefCnt;
     int		      unmapRefCnt;
 
     unsigned int initialViewportX;
     unsigned int initialViewportY;
 
+    Time initialTimestamp;
+    Bool initialTimestampSet;
+
     Bool placed;
     Bool minimized;
     Bool inShowDesktopMode;
     Bool shaded;
     Bool hidden;
 
+    unsigned int desktop;
+
     int pendingUnmaps;
+    int pendingMaps;
 
     char *startupId;
     char *resName;
@@ -1802,7 +2099,6 @@ struct _CompWindow {
 
     WindowPaintAttrib paint;
     WindowPaintAttrib lastPaint;
-    Bool	      scaled;
 
     CompWindowExtents input;
     CompWindowExtents output;
@@ -1844,6 +2140,9 @@ struct _CompWindow {
     int      texCoordSize;
     int      indexCount;
 
+    /* must be set by addWindowGeometry */
+    DrawWindowGeometryProc drawWindowGeometry;
+
     CompPrivate *privates;
 };
 
@@ -1891,6 +2190,18 @@ unsigned int
 getProtocols (CompDisplay *display,
 	      Window      id);
 
+unsigned int
+getWindowProp (CompDisplay  *display,
+	       Window	    id,
+	       Atom	    property,
+	       unsigned int defaultValue);
+
+void
+setWindowProp (CompDisplay  *display,
+	       Window       id,
+	       Atom	    property,
+	       unsigned int value);
+
 unsigned short
 getWindowProp32 (CompDisplay	*display,
 		 Window		id,
@@ -1932,8 +2243,10 @@ setWmState (CompDisplay *display,
 
 void
 setWindowFrameExtents (CompWindow	 *w,
-		       CompWindowExtents *input,
-		       CompWindowExtents *output);
+		       CompWindowExtents *input);
+
+void
+updateWindowOutputExtents (CompWindow *w);
 
 void
 updateWindowRegion (CompWindow *w);
@@ -1961,7 +2274,7 @@ mapWindow (CompWindow *w);
 void
 unmapWindow (CompWindow *w);
 
-void
+Bool
 bindWindow (CompWindow *w);
 
 void
@@ -2007,6 +2320,14 @@ void
 circulateWindow (CompWindow	 *w,
 		 XCirculateEvent *ce);
 
+void
+addWindowDamageRect (CompWindow *w,
+		     BoxPtr     rect);
+
+void
+getOutputExtentsForWindow (CompWindow	     *w,
+			   CompWindowExtents *output);
+
 void
 addWindowDamage (CompWindow *w);
 
@@ -2019,13 +2340,12 @@ damageWindowRect (CompWindow *w,
 		  BoxPtr     rect);
 
 void
-damageWindowRegion (CompWindow *w,
-		    Region     region);
-
-void
-setWindowScale (CompWindow *w,
-		float      xScale,
-		float      yScale);
+damageTransformedWindowRect (CompWindow *w,
+			     float	xScale,
+			     float	yScale,
+			     float	xTranslate,
+			     float	yTranslate,
+			     BoxPtr     rect);
 
 Bool
 focusWindow (CompWindow *w);
@@ -2143,9 +2463,18 @@ freeWindowIcons (CompWindow *w);
 int
 outputDeviceForWindow (CompWindow *w);
 
+Bool
+onCurrentDesktop (CompWindow *w);
+
+void
+setDesktopForWindow (CompWindow   *w,
+		     unsigned int desktop);
+
 
 /* plugin.c */
 
+#define HOME_PLUGINDIR ".compiz/plugins"
+
 typedef int (*GetVersionProc) (CompPlugin *plugin,
 			       int	  version);
 
@@ -2206,8 +2535,12 @@ typedef Bool (*LoadPluginProc) (CompPlugin *p,
 
 typedef void (*UnloadPluginProc) (CompPlugin *p);
 
+typedef char **(*ListPluginsProc) (char *path,
+				   int	*n);
+
 extern LoadPluginProc   loaderLoadPlugin;
 extern UnloadPluginProc loaderUnloadPlugin;
+extern ListPluginsProc  loaderListPlugins;
 
 struct _CompPlugin {
     CompPlugin       *next;
@@ -2268,6 +2601,9 @@ popPlugin (void);
 CompPlugin *
 getPlugins (void);
 
+char **
+availablePlugins (int *n);
+
 
 /* session.c */
 
@@ -2277,4 +2613,125 @@ initSession (char *smPrevClientId);
 void
 closeSession (void);
 
+
+/* fragment.c */
+
+#define MAX_FRAGMENT_FUNCTIONS 16
+
+struct _FragmentAttrib {
+    GLushort opacity;
+    GLushort brightness;
+    GLushort saturation;
+    int	     nTexture;
+    int	     function[MAX_FRAGMENT_FUNCTIONS];
+    int	     nFunction;
+    int	     nParam;
+};
+
+CompFunctionData *
+createFunctionData (void);
+
+void
+destroyFunctionData (CompFunctionData *data);
+
+Bool
+addTempHeaderOpToFunctionData (CompFunctionData *data,
+			       char		*name);
+
+Bool
+addParamHeaderOpToFunctionData (CompFunctionData *data,
+				char		 *name);
+
+Bool
+addAttribHeaderOpToFunctionData (CompFunctionData *data,
+				 char		  *name);
+
+#define COMP_FETCH_TARGET_2D   0
+#define COMP_FETCH_TARGET_RECT 1
+#define COMP_FETCH_TARGET_NUM  2
+
+Bool
+addFetchOpToFunctionData (CompFunctionData *data,
+			  char		   *dst,
+			  char		   *offset,
+			  int		   target);
+
+Bool
+addColorOpToFunctionData (CompFunctionData *data,
+			  char		   *dst,
+			  char		   *src);
+
+Bool
+addDataOpToFunctionData (CompFunctionData *data,
+			 char		  *str);
+
+Bool
+addBlendOpToFunctionData (CompFunctionData *data,
+			  char		   *str);
+
+int
+createFragmentFunction (CompScreen	 *s,
+			char		 *name,
+			CompFunctionData *data);
+
+void
+destroyFragmentFunction (CompScreen *s,
+			 int	    id);
+
+int
+getSaturateFragmentFunction (CompScreen  *s,
+			     CompTexture *texture,
+			     int	 param);
+
+int
+allocFragmentTextureUnits (FragmentAttrib *attrib,
+			   int		  nTexture);
+
+int
+allocFragmentParameters (FragmentAttrib *attrib,
+			 int		nParam);
+
+void
+addFragmentFunction (FragmentAttrib *attrib,
+		     int	    function);
+
+void
+initFragmentAttrib (FragmentAttrib	    *attrib,
+		    const WindowPaintAttrib *paint);
+
+Bool
+enableFragmentAttrib (CompScreen     *s,
+		      FragmentAttrib *attrib,
+		      Bool	     *blending);
+
+void
+disableFragmentAttrib (CompScreen     *s,
+		       FragmentAttrib *attrib);
+
+
+/* matrix.c */
+
+void
+matrixRotate (CompTransform *transform,
+	      float	    angle,
+	      float	    x,
+	      float	    y,
+	      float	    z);
+
+void
+matrixScale (CompTransform *transform,
+	     float	   x,
+	     float	   y,
+	     float	   z);
+
+void
+matrixTranslate (CompTransform *transform,
+		 float	       x,
+		 float	       y,
+		 float	       z);
+
+#ifdef  __cplusplus
+}
+#endif
+
 #endif
