@@ -30,7 +30,7 @@
 
 #define FADE_SPEED_DEFAULT    5.0f
 #define FADE_SPEED_MIN        0.1f
-#define FADE_SPEED_MAX       10.0f
+#define FADE_SPEED_MAX       25.0f
 #define FADE_SPEED_PRECISION  0.1f
 
 static char *winType[] = {
@@ -96,6 +96,7 @@ typedef struct _FadeWindow {
     int unmapCnt;
 
     Bool shaded;
+    Bool fadeOut;
 
     int steps;
 } FadeWindow;
@@ -261,6 +262,7 @@ fadeWindowStop (CompWindow *w)
 static Bool
 fadePaintWindow (CompWindow		 *w,
 		 const WindowPaintAttrib *attrib,
+		 const CompTransform	 *transform,
 		 Region			 region,
 		 unsigned int		 mask)
 {
@@ -273,98 +275,102 @@ fadePaintWindow (CompWindow		 *w,
     if (!w->screen->canDoSlightlySaturated)
 	fw->saturation = attrib->saturation;
 
-    if (fw->opacity    != attrib->opacity    ||
+    if (fw->destroyCnt			     ||
+	fw->unmapCnt			     ||
+	fw->opacity    != attrib->opacity    ||
 	fw->brightness != attrib->brightness ||
 	fw->saturation != attrib->saturation)
     {
-	GLint opacity;
-	GLint brightness;
-	GLint saturation;
+	WindowPaintAttrib fAttrib = *attrib;
 
-	opacity = fw->opacity;
-	if (attrib->opacity > fw->opacity)
+	if (fw->fadeOut)
+	    fAttrib.opacity = 0;
+
+	if (fw->steps)
 	{
-	    opacity = fw->opacity + fw->steps;
-	    if (opacity > attrib->opacity)
-		opacity = attrib->opacity;
-	}
-	else if (attrib->opacity < fw->opacity)
-	{
-	    if (w->type & CompWindowTypeUnknownMask)
-		opacity = fw->opacity - (fw->steps >> 1);
-	    else
-		opacity = fw->opacity - fw->steps;
+	    GLint opacity;
+	    GLint brightness;
+	    GLint saturation;
 
-	    if (opacity < attrib->opacity)
-		opacity = attrib->opacity;
-	}
+	    opacity = fw->opacity;
+	    if (fAttrib.opacity > fw->opacity)
+	    {
+		opacity = fw->opacity + fw->steps;
+		if (opacity > fAttrib.opacity)
+		    opacity = fAttrib.opacity;
+	    }
+	    else if (fAttrib.opacity < fw->opacity)
+	    {
+		if (w->type & CompWindowTypeUnknownMask)
+		    opacity = fw->opacity - (fw->steps >> 1);
+		else
+		    opacity = fw->opacity - fw->steps;
 
-	brightness = fw->brightness;
-	if (attrib->brightness > fw->brightness)
-	{
-	    brightness = fw->brightness + (fw->steps / 12);
-	    if (brightness > attrib->brightness)
-		brightness = attrib->brightness;
-	}
-	else if (attrib->brightness < fw->brightness)
-	{
-	    brightness = fw->brightness - (fw->steps / 12);
-	    if (brightness < attrib->brightness)
-		brightness = attrib->brightness;
-	}
+		if (opacity < fAttrib.opacity)
+		    opacity = fAttrib.opacity;
+	    }
 
-	saturation = fw->saturation;
-	if (attrib->saturation > fw->saturation)
-	{
-	    saturation = fw->saturation + (fw->steps / 6);
-	    if (saturation > attrib->saturation)
-		saturation = attrib->saturation;
-	}
-	else if (attrib->saturation < fw->saturation)
-	{
-	    saturation = fw->saturation - (fw->steps / 6);
-	    if (saturation < attrib->saturation)
-		saturation = attrib->saturation;
-	}
+	    brightness = fw->brightness;
+	    if (attrib->brightness > fw->brightness)
+	    {
+		brightness = fw->brightness + (fw->steps / 12);
+		if (brightness > attrib->brightness)
+		    brightness = attrib->brightness;
+	    }
+	    else if (attrib->brightness < fw->brightness)
+	    {
+		brightness = fw->brightness - (fw->steps / 12);
+		if (brightness < attrib->brightness)
+		    brightness = attrib->brightness;
+	    }
 
-	if (opacity > 0)
-	{
-	    WindowPaintAttrib fAttrib = *attrib;
+	    saturation = fw->saturation;
+	    if (attrib->saturation > fw->saturation)
+	    {
+		saturation = fw->saturation + (fw->steps / 6);
+		if (saturation > attrib->saturation)
+		    saturation = attrib->saturation;
+	    }
+	    else if (attrib->saturation < fw->saturation)
+	    {
+		saturation = fw->saturation - (fw->steps / 6);
+		if (saturation < attrib->saturation)
+		    saturation = attrib->saturation;
+	    }
 
-	    fAttrib.opacity    = opacity;
-	    fAttrib.brightness = brightness;
-	    fAttrib.saturation = saturation;
+	    fw->steps = 0;
 
-	    UNWRAP (fs, s, paintWindow);
-	    status = (*s->paintWindow) (w, &fAttrib, region, mask);
-	    WRAP (fs, s, paintWindow, fadePaintWindow);
-
-	    if (status)
+	    if (opacity > 0)
 	    {
 		fw->opacity    = opacity;
 		fw->brightness = brightness;
 		fw->saturation = saturation;
-		fw->steps      = 0;
 
 		if (opacity    != attrib->opacity    ||
 		    brightness != attrib->brightness ||
 		    saturation != attrib->saturation)
 		    addWindowDamage (w);
 	    }
-	}
-	else
-	{
-	    fw->opacity = 0;
+	    else
+	    {
+		fw->opacity = 0;
 
-	    fadeWindowStop (w);
-
-	    return (mask & PAINT_WINDOW_SOLID_MASK) ? FALSE : TRUE;
+		fadeWindowStop (w);
+	    }
 	}
+
+	fAttrib.opacity	   = fw->opacity;
+	fAttrib.brightness = fw->brightness;
+	fAttrib.saturation = fw->saturation;
+
+	UNWRAP (fs, s, paintWindow);
+	status = (*s->paintWindow) (w, &fAttrib, transform, region, mask);
+	WRAP (fs, s, paintWindow, fadePaintWindow);
     }
     else
     {
 	UNWRAP (fs, s, paintWindow);
-	status = (*s->paintWindow) (w, attrib, region, mask);
+	status = (*s->paintWindow) (w, attrib, transform, region, mask);
 	WRAP (fs, s, paintWindow, fadePaintWindow);
     }
 
@@ -466,14 +472,14 @@ fadeHandleEvent (CompDisplay *d,
 	    {
 		FADE_WINDOW (w);
 
-		w->paint.opacity = 0;
-
 		if (fw->opacity == 0xffff)
 		    fw->opacity = 0xfffe;
 
 		fw->destroyCnt++;
 		w->destroyRefCnt++;
 
+		fw->fadeOut = TRUE;
+
 		addWindowDamage (w);
 	    }
 
@@ -491,14 +497,14 @@ fadeHandleEvent (CompDisplay *d,
 
 	    if (!fw->shaded && w->texture->pixmap && (fs->wMask & w->type))
 	    {
-		w->paint.opacity = 0;
-
 		if (fw->opacity == 0xffff)
 		    fw->opacity = 0xfffe;
 
 		fw->unmapCnt++;
 		w->unmapRefCnt++;
 
+		fw->fadeOut = TRUE;
+
 		addWindowDamage (w);
 	    }
 
@@ -509,11 +515,6 @@ fadeHandleEvent (CompDisplay *d,
 	w = findWindowAtDisplay (d, event->xmap.window);
 	if (w)
 	{
-	    if (!(w->type & CompWindowTypeDesktopMask))
-		w->paint.opacity = getWindowProp32 (d, w->id,
-						    d->winOpacityAtom,
-						    OPAQUE);
-
 	    fadeWindowStop (w);
 
 	    if (w->state & CompWindowStateDisplayModalMask)
@@ -614,11 +615,13 @@ fadeDamageWindowRect (CompWindow *w,
     {
 	FADE_WINDOW (w);
 
+	fw->fadeOut = FALSE;
+
 	if (fw->shaded)
 	{
 	    fw->shaded = w->shaded;
 	}
-	else if ((fs->wMask & w->type) && fw->opacity == w->paint.opacity)
+	else if ((fs->wMask & w->type))
 	{
 	    fw->opacity = 0;
 	}
@@ -773,6 +776,7 @@ fadeInitWindow (CompPlugin *p,
     fw->destroyCnt = 0;
     fw->unmapCnt   = 0;
     fw->shaded     = w->shaded;
+    fw->fadeOut    = FALSE;
 
     w->privates[fs->windowPrivateIndex].ptr = fw;
 
@@ -794,6 +798,10 @@ fadeFiniWindow (CompPlugin *p,
     fadeRemoveDisplayModal (w->screen->display, w);
     fadeWindowStop (w);
 
+    w->paint.opacity    = w->opacity;
+    w->paint.brightness = w->brightness;
+    w->paint.saturation = w->saturation;
+
     free (fw);
 }
 
