@@ -590,6 +590,7 @@ minDonePaintScreen (CompScreen *s)
 static Bool
 minPaintScreen (CompScreen		*s,
 		const ScreenPaintAttrib *sAttrib,
+		const CompTransform	*transform,
 		Region		        region,
 		int			output,
 		unsigned int		mask)
@@ -602,7 +603,7 @@ minPaintScreen (CompScreen		*s,
 	mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;
 
     UNWRAP (ms, s, paintScreen);
-    status = (*s->paintScreen) (s, sAttrib, region, output, mask);
+    status = (*s->paintScreen) (s, sAttrib, transform, region, output, mask);
     WRAP (ms, s, paintScreen, minPaintScreen);
 
     return status;
@@ -611,6 +612,7 @@ minPaintScreen (CompScreen		*s,
 static Bool
 minPaintWindow (CompWindow		*w,
 		const WindowPaintAttrib *attrib,
+		const CompTransform	*transform,
 		Region			region,
 		unsigned int		mask)
 {
@@ -622,28 +624,42 @@ minPaintWindow (CompWindow		*w,
 
     if (mw->adjust)
     {
-	WindowPaintAttrib mAttrib;
+	FragmentAttrib fragment;
+	CompTransform  wTransform = *transform;
 
 	UNWRAP (ms, s, paintWindow);
-	status = (*s->paintWindow) (w, attrib, region,
+	status = (*s->paintWindow) (w, attrib, transform, region,
 				    mask | PAINT_WINDOW_NO_CORE_INSTANCE_MASK);
 	WRAP (ms, s, paintWindow, minPaintWindow);
 
-	mAttrib = w->lastPaint;
+	initFragmentAttrib (&fragment, &w->lastPaint);
 
-	mAttrib.xScale = mw->xScale;
-	mAttrib.yScale = mw->yScale;
+	if (w->alpha || fragment.opacity != OPAQUE)
+	    mask |= PAINT_WINDOW_TRANSLUCENT_MASK;
 
-	mAttrib.xTranslate = mw->tx;
-	mAttrib.yTranslate = mw->ty;
+	matrixTranslate (&wTransform, w->attrib.x, w->attrib.y, 0.0f);
+	matrixScale (&wTransform, mw->xScale, mw->yScale, 0.0f);
+	matrixTranslate (&wTransform,
+			 mw->tx / mw->xScale - w->attrib.x,
+			 mw->ty / mw->yScale - w->attrib.y,
+			 0.0f);
 
-	(*s->drawWindow) (w, &mAttrib, region,
+	glPushMatrix ();
+	glLoadMatrixf (wTransform.m);
+
+	(*s->drawWindow) (w, &wTransform, &fragment, region,
 			  mask | PAINT_WINDOW_TRANSFORMED_MASK);
+
+	glPopMatrix ();
     }
     else
     {
+	/* no core instance from windows that have been animated */
+	if (mw->state == IconicState)
+	    mask |= PAINT_WINDOW_NO_CORE_INSTANCE_MASK;
+
 	UNWRAP (ms, s, paintWindow);
-	status = (*s->paintWindow) (w, attrib, region, mask);
+	status = (*s->paintWindow) (w, attrib, transform, region, mask);
 	WRAP (ms, s, paintWindow, minPaintWindow);
     }
 
@@ -714,10 +730,10 @@ minHandleEvent (CompDisplay *d,
 		}
 		else if (!w->invisible && (ms->wMask & w->type))
 		{
-		    mw->newState = IconicState;
-
 		    if (minGetWindowIconGeometry (w, &mw->icon))
 		    {
+			mw->newState = IconicState;
+
 			mw->xScale = w->paint.xScale;
 			mw->yScale = w->paint.yScale;
 			mw->tx	   = w->attrib.x - w->serverX;
@@ -739,10 +755,6 @@ minHandleEvent (CompDisplay *d,
 
 			addWindowDamage (w);
 		    }
-		    else
-		    {
-			mw->state = mw->newState;
-		    }
 		}
 	    }
 	    else  /* X -> Withdrawn */
@@ -791,6 +803,8 @@ minDamageWindowRect (CompWindow *w,
     {
 	if (mw->state == IconicState)
 	{
+	    mw->state = NormalState;
+
 	    if (!w->invisible	      &&
 		(ms->wMask & w->type) &&
 		minGetWindowIconGeometry (w, &mw->icon))
@@ -817,7 +831,7 @@ minDamageWindowRect (CompWindow *w,
 		XOffsetRegion (mw->region, -w->attrib.x, -w->attrib.y);
 
 		/* bind pixmap here so we have something to unshade with */
-		if (!w->texture->pixmap)
+		if (!w->texture->pixmap && !w->bindFailed)
 		    bindWindow (w);
 
 		ms->moreAdjust = TRUE;
