@@ -29,6 +29,7 @@
 #include <math.h>
 #include <sys/time.h>
 
+#include <X11/Xatom.h>
 #include <X11/cursorfont.h>
 
 #include <compiz.h>
@@ -42,8 +43,6 @@
 #define SCALE_SPACING_MIN     0
 #define SCALE_SPACING_MAX     250
 
-#define SCALE_SLOPPY_FOCUS_DEFAULT TRUE
-
 #define SCALE_INITIATE_KEY_DEFAULT       "Up"
 #define SCALE_INITIATE_MODIFIERS_DEFAULT (ControlMask | CompAltMask)
 
@@ -98,6 +97,10 @@ static IconOverlay iconOverlay[] = {
 #define N_ICON_TYPE (sizeof (iconOverlayString) / sizeof (iconOverlayString[0]))
 #define SCALE_ICON_DEFAULT (iconOverlayString[1])
 
+#define SCALE_HOVER_TIME_DEFAULT 750
+#define SCALE_HOVER_TIME_MIN     50
+#define SCALE_HOVER_TIME_MAX     10000
+
 static int displayPrivateIndex;
 
 typedef struct _ScaleSlot {
@@ -106,8 +109,11 @@ typedef struct _ScaleSlot {
     float scale;
 } ScaleSlot;
 
-#define SCALE_DISPLAY_OPTION_INITIATE 0
-#define SCALE_DISPLAY_OPTION_NUM      1
+#define SCALE_DISPLAY_OPTION_INITIATE        0
+#define SCALE_DISPLAY_OPTION_INITIATE_ALL    1
+#define SCALE_DISPLAY_OPTION_INITIATE_GROUP  2
+#define SCALE_DISPLAY_OPTION_INITIATE_OUTPUT 3
+#define SCALE_DISPLAY_OPTION_NUM             4
 
 typedef struct _ScaleDisplay {
     int		    screenPrivateIndex;
@@ -116,19 +122,28 @@ typedef struct _ScaleDisplay {
     CompOption opt[SCALE_DISPLAY_OPTION_NUM];
 
     unsigned int lastActiveNum;
+    Window       lastActiveWindow;
+    Window       selectedWindow;
     KeyCode	 leftKeyCode, rightKeyCode, upKeyCode, downKeyCode;
 } ScaleDisplay;
 
 #define SCALE_SCREEN_OPTION_SPACING      0
-#define SCALE_SCREEN_OPTION_SLOPPY_FOCUS 1
-#define SCALE_SCREEN_OPTION_SPEED	 2
-#define SCALE_SCREEN_OPTION_TIMESTEP	 3
-#define SCALE_SCREEN_OPTION_WINDOW_TYPE  4
-#define SCALE_SCREEN_OPTION_DARKEN_BACK  5
-#define SCALE_SCREEN_OPTION_OPACITY      6
-#define SCALE_SCREEN_OPTION_ICON         7
+#define SCALE_SCREEN_OPTION_SPEED	 1
+#define SCALE_SCREEN_OPTION_TIMESTEP	 2
+#define SCALE_SCREEN_OPTION_WINDOW_TYPE  3
+#define SCALE_SCREEN_OPTION_DARKEN_BACK  4
+#define SCALE_SCREEN_OPTION_OPACITY      5
+#define SCALE_SCREEN_OPTION_ICON         6
+#define SCALE_SCREEN_OPTION_HOVER_TIME   7
 #define SCALE_SCREEN_OPTION_NUM          8
 
+typedef enum {
+    ScaleTypeNormal = 0,
+    ScaleTypeOutput,
+    ScaleTypeGroup,
+    ScaleTypeAll
+} ScaleType;
+
 typedef struct _ScaleScreen {
     int windowPrivateIndex;
 
@@ -147,7 +162,12 @@ typedef struct _ScaleScreen {
 
     unsigned int wMask;
 
-    int grabIndex;
+    Bool grab;
+    int  grabIndex;
+
+    Window dndTarget;
+
+    CompTimeoutHandle hoverHandle;
 
     int state;
     int moreAdjust;
@@ -167,6 +187,10 @@ typedef struct _ScaleScreen {
     GLushort opacity;
 
     IconOverlay iconOverlay;
+
+    ScaleType type;
+
+    Window clientLeader;
 } ScaleScreen;
 
 typedef struct _ScaleWindow {
@@ -180,6 +204,8 @@ typedef struct _ScaleWindow {
     GLfloat tx, ty;
     float   delta;
     Bool    adjust;
+
+    float lastThumbOpacity;
 } ScaleWindow;
 
 
@@ -238,10 +264,6 @@ scaleSetScreenOption (CompScreen      *screen,
 	    return TRUE;
 	}
 	break;
-    case SCALE_SCREEN_OPTION_SLOPPY_FOCUS:
-	if (compSetBoolOption (o, value))
-	    return TRUE;
-	break;
     case SCALE_SCREEN_OPTION_SPEED:
 	if (compSetFloatOption (o, value))
 	{
@@ -292,6 +314,9 @@ scaleSetScreenOption (CompScreen      *screen,
 	    }
 	}
 	break;
+    case SCALE_SCREEN_OPTION_HOVER_TIME:
+	if (compSetIntOption (o, value))
+	    return TRUE;
     default:
 	break;
     }
@@ -314,13 +339,6 @@ scaleScreenInitOptions (ScaleScreen *ss)
     o->rest.i.min = SCALE_SPACING_MIN;
     o->rest.i.max = SCALE_SPACING_MAX;
 
-    o = &ss->opt[SCALE_SCREEN_OPTION_SLOPPY_FOCUS];
-    o->name	  = "sloppy_focus";
-    o->shortDesc  = N_("Sloppy Focus");
-    o->longDesc   = N_("Focus window when mouse moves over them");
-    o->type	  = CompOptionTypeBool;
-    o->value.b	  = SCALE_SLOPPY_FOCUS_DEFAULT;
-
     o = &ss->opt[SCALE_SCREEN_OPTION_SPEED];
     o->name		= "speed";
     o->shortDesc	= N_("Speed");
@@ -380,6 +398,28 @@ scaleScreenInitOptions (ScaleScreen *ss)
     o->value.s	      = strdup (SCALE_ICON_DEFAULT);
     o->rest.s.string  = iconOverlayString;
     o->rest.s.nString = N_ICON_TYPE;
+
+    o = &ss->opt[SCALE_SCREEN_OPTION_HOVER_TIME];
+    o->name	  = "hover_time";
+    o->shortDesc  = N_("Hover Time");
+    o->longDesc	  = N_("Time (in ms) before scale mode is terminated when "
+		       "hovering over a window");
+    o->type	  = CompOptionTypeInt;
+    o->value.i    = SCALE_HOVER_TIME_DEFAULT;
+    o->rest.i.min = SCALE_HOVER_TIME_MIN;
+    o->rest.i.max = SCALE_HOVER_TIME_MAX;
+}
+
+static Bool
+isNeverScaleWin (CompWindow *w)
+{
+    if (w->attrib.override_redirect)
+	return TRUE;
+
+    if (w->wmType & (CompWindowTypeDockMask | CompWindowTypeDesktopMask))
+	return TRUE;
+
+    return FALSE;
 }
 
 static Bool
@@ -387,9 +427,15 @@ isScaleWin (CompWindow *w)
 {
     SCALE_SCREEN (w->screen);
 
-    if (!(*w->screen->focusWindow) (w))
+    if (isNeverScaleWin (w))
 	return FALSE;
 
+    if (!ss->type || ss->type == ScaleTypeOutput)
+    {
+	if (!(*w->screen->focusWindow) (w))
+	    return FALSE;
+    }
+
     if (!(ss->wMask & w->type))
 	return FALSE;
 
@@ -399,12 +445,29 @@ isScaleWin (CompWindow *w)
     if (w->state & CompWindowStateShadedMask)
 	return FALSE;
 
+    if (!w->mapNum || w->attrib.map_state != IsViewable)
+	return FALSE;
+
+    switch (ss->type) {
+    case ScaleTypeGroup:
+	if (ss->clientLeader != w->clientLeader &&
+	    ss->clientLeader != w->id)
+	    return FALSE;
+	break;
+    case ScaleTypeOutput:
+	if (outputDeviceForWindow(w) != w->screen->currentOutputDev)
+	    return FALSE;
+    default:
+	break;
+    }
+
     return TRUE;
 }
 
 static Bool
 scalePaintWindow (CompWindow		  *w,
 		  const WindowPaintAttrib *attrib,
+		  const CompTransform	  *transform,
 		  Region		  region,
 		  unsigned int		  mask)
 {
@@ -413,35 +476,76 @@ scalePaintWindow (CompWindow		  *w,
 
     SCALE_SCREEN (s);
 
-    if (ss->grabIndex)
+    if (ss->state != SCALE_STATE_NONE)
     {
 	WindowPaintAttrib sAttrib = *attrib;
+	Bool		  scaled = FALSE;
 
 	SCALE_WINDOW (w);
 
 	if (sw->adjust || sw->slot)
 	{
-	    mask |= PAINT_WINDOW_TRANSFORMED_MASK;
+	    SCALE_DISPLAY (s->display);
 
-	    if (w->id	    != s->display->activeWindow &&
-		ss->opacity != OPAQUE			&&
+	    if (w->id	    != sd->selectedWindow &&
+		ss->opacity != OPAQUE		  &&
 		ss->state   != SCALE_STATE_IN)
 	    {
 		/* modify opacity of windows that are not active */
 		sAttrib.opacity = (sAttrib.opacity * ss->opacity) >> 16;
 	    }
+
+	    scaled = TRUE;
+
+	    mask |= PAINT_WINDOW_NO_CORE_INSTANCE_MASK;
 	}
-	else if (ss->darkenBack && ss->state != SCALE_STATE_IN)
+	else if (ss->state != SCALE_STATE_IN)
 	{
-	    /* modify brightness of the other windows */
-	    sAttrib.brightness = sAttrib.brightness / 2;
+	    if (ss->darkenBack)
+	    {
+		/* modify brightness of the other windows */
+		sAttrib.brightness = sAttrib.brightness / 2;
+	    }
+
+	    /* hide windows on this output that are not in scale mode */
+	    if (!isNeverScaleWin (w) &&
+		outputDeviceForWindow (w) == s->currentOutputDev)
+	    {
+		sAttrib.opacity = 0;
+	    }
 	}
 
 	UNWRAP (ss, s, paintWindow);
-	status = (*s->paintWindow) (w, &sAttrib, region, mask);
+	status = (*s->paintWindow) (w, &sAttrib, transform, region, mask);
 	WRAP (ss, s, paintWindow, scalePaintWindow);
 
-	if ((ss->iconOverlay != ScaleIconNone) && (sw->adjust || sw->slot))
+	if (scaled)
+	{
+	    FragmentAttrib fragment;
+	    CompTransform  wTransform = *transform;
+
+	    initFragmentAttrib (&fragment, &w->lastPaint);
+
+	    if (w->alpha || fragment.opacity != OPAQUE)
+		mask |= PAINT_WINDOW_TRANSLUCENT_MASK;
+
+	    matrixTranslate (&wTransform, w->attrib.x, w->attrib.y, 0.0f);
+	    matrixScale (&wTransform, sw->scale, sw->scale, 0.0f);
+	    matrixTranslate (&wTransform,
+			     sw->tx / sw->scale - w->attrib.x,
+			     sw->ty / sw->scale - w->attrib.y,
+			     0.0f);
+
+	    glPushMatrix ();
+	    glLoadMatrixf (wTransform.m);
+
+	    (*s->drawWindow) (w, &wTransform, &fragment, region,
+			      mask | PAINT_WINDOW_TRANSFORMED_MASK);
+
+	    glPopMatrix ();
+	}
+
+	if ((ss->iconOverlay != ScaleIconNone) && scaled)
 	{
 	    CompIcon *icon;
 
@@ -451,11 +555,12 @@ scalePaintWindow (CompWindow		  *w,
 
 	    if (icon && (icon->texture.name || iconToTexture (w->screen, icon)))
 	    {
-		REGION	    iconReg;
-		CompMatrix  matrix;
-		int	    wx, wy, width, height;
-		int	    scaledWinWidth, scaledWinHeight;
-		float	    ds;
+		REGION iconReg;
+		float  scale;
+		float  x, y;
+		int    width, height;
+		int    scaledWinWidth, scaledWinHeight;
+		float  ds;
 
 		scaledWinWidth  = w->width  * sw->scale;
 		scaledWinHeight = w->height * sw->scale;
@@ -463,47 +568,47 @@ scalePaintWindow (CompWindow		  *w,
 		switch (ss->iconOverlay) {
 		case ScaleIconNone:
 		case ScaleIconEmblem:
-		    sAttrib.xScale = sAttrib.yScale = 1.0f;
+		    scale = 1.0f;
 		    break;
 		case ScaleIconBig:
-		    sAttrib.opacity /= 3;
-		    sAttrib.xScale = sAttrib.yScale =
-			MIN (((float) scaledWinWidth / (float) icon->width),
-			     ((float) scaledWinHeight / (float) icon->height));
 		default:
+		    sAttrib.opacity /= 3;
+		    scale = MIN (((float) scaledWinWidth / icon->width),
+				 ((float) scaledWinHeight / icon->height));
 		    break;
 		}
 
-		width  = icon->width  * sAttrib.xScale;
-		height = icon->height * sAttrib.yScale;
+		width  = icon->width  * scale;
+		height = icon->height * scale;
 
 		switch (ss->iconOverlay) {
 		case ScaleIconNone:
 		case ScaleIconEmblem:
-		    wx = w->attrib.x + (w->width  * sw->scale) - icon->width;
-		    wy = w->attrib.y + (w->height * sw->scale) - icon->height;
+		    x = w->attrib.x + scaledWinWidth - icon->width;
+		    y = w->attrib.y + scaledWinHeight - icon->height;
 		    break;
 		case ScaleIconBig:
-		    wx = w->attrib.x + ((scaledWinWidth - width) / 2) /
-			sAttrib.xScale;
-		    wy = w->attrib.y + ((scaledWinHeight - height) / 2) /
-			sAttrib.yScale;
-		    break;
 		default:
-		    wx = wy = 0;
+		    x = w->attrib.x + scaledWinWidth / 2 - width / 2;
+		    y = w->attrib.y + scaledWinHeight / 2 - height / 2;
 		    break;
 		}
 
+		x += sw->tx;
+		y += sw->ty;
+
 		if (sw->slot)
 		{
 		    sw->delta =
-			fabs (sw->slot->x1 - w->serverX) +
-			fabs (sw->slot->y1 - w->serverY) +
+			fabs (sw->slot->x1 - w->attrib.x) +
+			fabs (sw->slot->y1 - w->attrib.y) +
 			fabs (1.0f - sw->slot->scale) * 500.0f;
 		}
 
 		if (sw->delta)
 		{
+		    float o;
+
 		    ds =
 			fabs (sw->tx) +
 			fabs (sw->ty) +
@@ -512,40 +617,67 @@ scalePaintWindow (CompWindow		  *w,
 		    if (ds > sw->delta)
 			ds = sw->delta;
 
-		    sAttrib.opacity = (ds * sAttrib.opacity) / sw->delta;
+		    o = ds / sw->delta;
+
+		    if (sw->slot)
+		    {
+			if (o < sw->lastThumbOpacity)
+			    o = sw->lastThumbOpacity;
+		    }
+		    else
+		    {
+			if (o > sw->lastThumbOpacity)
+			    o = 0.0f;
+		    }
+
+		    sw->lastThumbOpacity = o;
+
+		    sAttrib.opacity = sAttrib.opacity * o;
 		}
 
-		mask |= PAINT_WINDOW_TRANSLUCENT_MASK;
+		mask |= PAINT_WINDOW_BLEND_MASK;
 
 		iconReg.rects    = &iconReg.extents;
 		iconReg.numRects = 1;
 
-		iconReg.extents.x1 = wx;
-		iconReg.extents.y1 = wy;
+		iconReg.extents.x1 = 0;
+		iconReg.extents.y1 = 0;
 		iconReg.extents.x2 = iconReg.extents.x1 + width;
 		iconReg.extents.y2 = iconReg.extents.y1 + height;
 
-		matrix = icon->texture.matrix;
-		matrix.x0 -= wx * icon->texture.matrix.xx;
-		matrix.y0 -= wy * icon->texture.matrix.yy;
-
-		w->vCount = 0;
+		w->vCount = w->indexCount = 0;
 		if (iconReg.extents.x1 < iconReg.extents.x2 &&
 		    iconReg.extents.y1 < iconReg.extents.y2)
-		    (*w->screen->addWindowGeometry) (w, &matrix, 1, &iconReg,
-						     &iconReg);
+		    (*w->screen->addWindowGeometry) (w,
+						     &icon->texture.matrix, 1,
+						     &iconReg, &iconReg);
 
 		if (w->vCount)
+		{
+		    FragmentAttrib fragment;
+		    CompTransform  wTransform = *transform;
+
+		    initFragmentAttrib (&fragment, &sAttrib);
+
+		    matrixScale (&wTransform, scale, scale, 1.0f);
+		    matrixTranslate (&wTransform, x / scale, y / scale, 0.0f);
+
+		    glPushMatrix ();
+		    glLoadMatrixf (wTransform.m);
+
 		    (*w->screen->drawWindowTexture) (w,
-						     &icon->texture, &sAttrib,
+						     &icon->texture, &fragment,
 						     mask);
+
+		    glPopMatrix ();
+		}
 	    }
 	}
     }
     else
     {
 	UNWRAP (ss, s, paintWindow);
-	status = (*s->paintWindow) (w, attrib, region, mask);
+	status = (*s->paintWindow) (w, attrib, transform, region, mask);
 	WRAP (ss, s, paintWindow, scalePaintWindow);
     }
 
@@ -570,6 +702,7 @@ static void
 layoutSlots (CompScreen *s)
 {
     int i, j, x, y, width, height, lines, n;
+    int x1, y1, x2, y2;
 
     SCALE_SCREEN (s);
 
@@ -577,16 +710,27 @@ layoutSlots (CompScreen *s)
 
     lines = sqrt (ss->nWindows + 1);
 
-    y      = s->workArea.y + ss->spacing;
-    height = (s->workArea.height - (lines + 1) * ss->spacing) / lines;
+    getCurrentOutputExtents (s, &x1, &y1, &x2, &y2);
+
+    if (s->workArea.x > x1)
+	x1 = s->workArea.x;
+    if (s->workArea.x + s->workArea.width < x2)
+	x2 = s->workArea.x + s->workArea.width;
+    if (s->workArea.y > y1)
+	y1 = s->workArea.y;
+    if (s->workArea.y + s->workArea.height < y2)
+	y2 = s->workArea.y + s->workArea.height;
+
+    y      = y1 + ss->spacing;
+    height = ((y2 - y1) - (lines + 1) * ss->spacing) / lines;
 
     for (i = 0; i < lines; i++)
     {
 	n = MIN (ss->nWindows - ss->nSlots,
 		 ceilf ((float) ss->nWindows / lines));
 
-	x     = s->workArea.x + ss->spacing;
-	width = (s->workArea.width - (n + 1) * ss->spacing) / n;
+	x     = x1 + ss->spacing;
+	width = ((x2 - x1) - (n + 1) * ss->spacing) / n;
 
 	for (j = 0; j < n; j++)
 	{
@@ -683,13 +827,13 @@ fillInWindows (CompScreen *s)
 
 	    sw->slot->scale = MIN (MIN (sx, sy), 1.0f);
 
-	    sx = w->width  * sw->slot->scale;
-	    sy = w->height * sw->slot->scale;
+	    sx = width  * sw->slot->scale;
+	    sy = height * sw->slot->scale;
 	    cx = (sw->slot->x1 + sw->slot->x2) / 2;
 	    cy = (sw->slot->y1 + sw->slot->y2) / 2;
 
-	    cx += (w->input.left - w->input.right)  * sw->slot->scale;
-	    cy += (w->input.top  - w->input.bottom) * sw->slot->scale;
+	    cx += w->input.left * sw->slot->scale;
+	    cy += w->input.top  * sw->slot->scale;
 
 	    sw->slot->x1 = cx - sx / 2;
 	    sw->slot->y1 = cy - sy / 2;
@@ -698,6 +842,8 @@ fillInWindows (CompScreen *s)
 
 	    sw->slot->filled = TRUE;
 
+	    sw->lastThumbOpacity = 0.0f;
+
 	    sw->adjust = TRUE;
 	}
     }
@@ -709,7 +855,6 @@ static Bool
 layoutThumbs (CompScreen *s)
 {
     CompWindow *w;
-    int	       i;
 
     SCALE_SCREEN (s);
 
@@ -767,14 +912,6 @@ layoutThumbs (CompScreen *s)
 
     } while (fillInWindows (s));
 
-    for (i = 0; i < ss->nWindows; i++)
-    {
-	SCALE_WINDOW (ss->windows[i]);
-
-	if (sw->slot)
-	    sw->adjust = TRUE;
-    }
-
     return TRUE;
 }
 
@@ -794,12 +931,12 @@ adjustScaleVelocity (CompWindow *w)
     }
     else
     {
-	x1 = w->serverX;
-	y1 = w->serverY;
+	x1 = w->attrib.x;
+	y1 = w->attrib.y;
 	scale = 1.0f;
     }
 
-    dx = x1 - (w->serverX + sw->tx);
+    dx = x1 - (w->attrib.x + sw->tx);
 
     adjust = dx * 0.15f;
     amount = fabs (dx) * 1.5f;
@@ -810,7 +947,7 @@ adjustScaleVelocity (CompWindow *w)
 
     sw->xVelocity = (amount * sw->xVelocity + adjust) / (amount + 1.0f);
 
-    dy = y1 - (w->serverY + sw->ty);
+    dy = y1 - (w->attrib.y + sw->ty);
 
     adjust = dy * 0.15f;
     amount = fabs (dy) * 1.5f;
@@ -838,8 +975,8 @@ adjustScaleVelocity (CompWindow *w)
 	fabs (ds) < 0.001f && fabs (sw->scaleVelocity) < 0.002f)
     {
 	sw->xVelocity = sw->yVelocity = sw->scaleVelocity = 0.0f;
-	sw->tx = x1 - w->serverX;
-	sw->ty = y1 - w->serverY;
+	sw->tx = x1 - w->attrib.x;
+	sw->ty = y1 - w->attrib.y;
 	sw->scale = scale;
 
 	return 0;
@@ -851,6 +988,7 @@ adjustScaleVelocity (CompWindow *w)
 static Bool
 scalePaintScreen (CompScreen		  *s,
 		  const ScreenPaintAttrib *sAttrib,
+		  const CompTransform	  *transform,
 		  Region		  region,
 		  int			  output,
 		  unsigned int		  mask)
@@ -859,11 +997,11 @@ scalePaintScreen (CompScreen		  *s,
 
     SCALE_SCREEN (s);
 
-    if (ss->grabIndex)
+    if (ss->state != SCALE_STATE_NONE)
 	mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;
 
     UNWRAP (ss, s, paintScreen);
-    status = (*s->paintScreen) (s, sAttrib, region, output, mask);
+    status = (*s->paintScreen) (s, sAttrib, transform, region, output, mask);
     WRAP (ss, s, paintScreen, scalePaintScreen);
 
     return status;
@@ -875,10 +1013,10 @@ scalePreparePaintScreen (CompScreen *s,
 {
     SCALE_SCREEN (s);
 
-    if (ss->grabIndex && ss->state != SCALE_STATE_WAIT)
+    if (ss->state != SCALE_STATE_NONE && ss->state != SCALE_STATE_WAIT)
     {
 	CompWindow *w;
-	int        steps, dx, dy;
+	int        steps;
 	float      amount, chunk;
 
 	amount = msSinceLastPaint * 0.05f * ss->speed;
@@ -903,13 +1041,6 @@ scalePreparePaintScreen (CompScreen *s,
 		    sw->tx += sw->xVelocity * chunk;
 		    sw->ty += sw->yVelocity * chunk;
 		    sw->scale += sw->scaleVelocity * chunk;
-
-		    dx = (w->serverX + sw->tx) - w->attrib.x;
-		    dy = (w->serverY + sw->ty) - w->attrib.y;
-
-		    moveWindow (w, dx, dy, FALSE, FALSE);
-
-		    (*s->setWindowScale) (w, sw->scale, sw->scale);
 		}
 	    }
 
@@ -928,7 +1059,7 @@ scaleDonePaintScreen (CompScreen *s)
 {
     SCALE_SCREEN (s);
 
-    if (ss->grabIndex)
+    if (ss->state != SCALE_STATE_NONE)
     {
 	if (ss->moreAdjust)
 	{
@@ -937,10 +1068,7 @@ scaleDonePaintScreen (CompScreen *s)
 	else
 	{
 	    if (ss->state == SCALE_STATE_IN)
-	    {
-		removeScreenGrab (s, ss->grabIndex, 0);
-		ss->grabIndex = 0;
-	    }
+		ss->state = SCALE_STATE_NONE;
 	    else if (ss->state == SCALE_STATE_OUT)
 		ss->state = SCALE_STATE_WAIT;
 	}
@@ -970,6 +1098,11 @@ scaleCheckForWindowAt (CompScreen *s,
 	    x2 = w->attrib.x + (w->width  + w->input.right)  * sw->scale;
 	    y2 = w->attrib.y + (w->height + w->input.bottom) * sw->scale;
 
+	    x1 += sw->tx;
+	    y1 += sw->ty;
+	    x2 += sw->tx;
+	    y2 += sw->ty;
+
 	    if (x1 <= x && y1 <= y && x2 > x && y2 > y)
 		return w;
 	}
@@ -978,6 +1111,57 @@ scaleCheckForWindowAt (CompScreen *s,
     return 0;
 }
 
+static void
+sendViewportMoveRequest (CompScreen *s,
+			 int	    x,
+			 int	    y)
+{
+    XEvent xev;
+
+    xev.xclient.type    = ClientMessage;
+    xev.xclient.display = s->display->display;
+    xev.xclient.format  = 32;
+
+    xev.xclient.message_type = s->display->desktopViewportAtom;
+    xev.xclient.window	     = s->root;
+
+    xev.xclient.data.l[0] = x;
+    xev.xclient.data.l[1] = y;
+    xev.xclient.data.l[2] = 0;
+    xev.xclient.data.l[3] = 0;
+    xev.xclient.data.l[4] = 0;
+
+    XSendEvent (s->display->display,
+		s->root,
+		FALSE,
+		SubstructureRedirectMask | SubstructureNotifyMask,
+		&xev);
+}
+
+static void
+sendDndStatusMessage (CompScreen *s,
+		      Window	 source)
+{
+    XEvent xev;
+
+    SCALE_SCREEN (s);
+
+    xev.xclient.type    = ClientMessage;
+    xev.xclient.display = s->display->display;
+    xev.xclient.format  = 32;
+
+    xev.xclient.message_type = s->display->xdndStatusAtom;
+    xev.xclient.window	     = source;
+
+    xev.xclient.data.l[0] = ss->dndTarget;
+    xev.xclient.data.l[1] = 2;
+    xev.xclient.data.l[2] = 0;
+    xev.xclient.data.l[3] = 0;
+    xev.xclient.data.l[4] = None;
+
+    XSendEvent (s->display->display, source, FALSE, 0, &xev);
+}
+
 static Bool
 scaleTerminate (CompDisplay     *d,
 		CompAction      *action,
@@ -999,14 +1183,20 @@ scaleTerminate (CompDisplay     *d,
 	if (xid && s->root != xid)
 	    continue;
 
-	if (ss->grabIndex)
+	if (ss->grab)
 	{
-	    if (ss->state == SCALE_STATE_NONE)
+	    if (ss->grabIndex)
 	    {
 		removeScreenGrab (s, ss->grabIndex, 0);
 		ss->grabIndex = 0;
 	    }
-	    else
+
+	    if (ss->dndTarget)
+		XUnmapWindow (d->display, ss->dndTarget);
+
+	    ss->grab = FALSE;
+
+	    if (ss->state != SCALE_STATE_NONE)
 	    {
 		CompWindow *w;
 
@@ -1021,18 +1211,119 @@ scaleTerminate (CompDisplay     *d,
 		    }
 		}
 
+		if (ss->state != SCALE_STATE_IN)
+		{
+		    w = findWindowAtScreen (s, sd->lastActiveWindow);
+		    if (w)
+		    {
+			int x, y;
+
+			activateWindow (w);
+
+			defaultViewportForWindow (w, &x, &y);
+
+			if (x != s->x || y != s->y)
+			    sendViewportMoveRequest (s,
+						     x * s->width,
+						     y * s->height);
+		    }
+		}
+
 		ss->state = SCALE_STATE_IN;
 
 		damageScreen (s);
 	    }
 
-	    sd->lastActiveNum = None;
+	    sd->lastActiveNum = 0;
 	}
     }
 
     return FALSE;
 }
 
+static Bool
+scaleEnsureDndRedirectWindow (CompScreen *s)
+{
+    SCALE_SCREEN (s);
+
+    if (!ss->dndTarget)
+    {
+	XSetWindowAttributes attr;
+	long		     xdndVersion = 3;
+
+	attr.override_redirect = TRUE;
+
+	ss->dndTarget = XCreateWindow (s->display->display,
+				       s->root,
+				       0, 0, 1, 1, 0,
+				       CopyFromParent,
+				       InputOnly,
+				       CopyFromParent,
+				       CWOverrideRedirect, &attr);
+
+	XChangeProperty (s->display->display, ss->dndTarget,
+			 s->display->xdndAwareAtom,
+			 XA_ATOM, 32, PropModeReplace,
+			 (unsigned char *) &xdndVersion, 1);
+    }
+
+    XMoveResizeWindow (s->display->display, ss->dndTarget,
+		       0, 0, s->width, s->height);
+    XMapRaised (s->display->display, ss->dndTarget);
+
+    return TRUE;
+}
+
+static Bool
+scaleInitiateCommon (CompScreen      *s,
+		     CompAction      *action,
+		     CompActionState state,
+		     CompOption      *option,
+		     int	     nOption)
+{
+    SCALE_DISPLAY (s->display);
+    SCALE_SCREEN (s);
+
+    if (otherScreenGrabExist (s, "scale", 0))
+	return FALSE;
+
+    if (!layoutThumbs (s))
+	return FALSE;
+
+    if (state & CompActionStateInitEdgeDnd)
+    {
+	if (scaleEnsureDndRedirectWindow (s))
+	    ss->grab = TRUE;
+    }
+    else if (!ss->grabIndex)
+    {
+	ss->grabIndex = pushScreenGrab (s, ss->cursor, "scale");
+	if (ss->grabIndex)
+	    ss->grab = TRUE;
+    }
+
+    if (ss->grab)
+    {
+	if (!sd->lastActiveNum)
+	    sd->lastActiveNum = s->activeNum - 1;
+
+	sd->lastActiveWindow = s->display->activeWindow;
+	sd->selectedWindow   = s->display->activeWindow;
+
+	ss->state = SCALE_STATE_OUT;
+
+	damageScreen (s);
+    }
+
+    if (state & CompActionStateInitButton)
+	action->state |= CompActionStateTermButton;
+
+    if (state & CompActionStateInitKey)
+	action->state |= CompActionStateTermKey;
+
+    return FALSE;
+}
+
 static Bool
 scaleInitiate (CompDisplay     *d,
 	       CompAction      *action,
@@ -1048,48 +1339,135 @@ scaleInitiate (CompDisplay     *d,
     s = findScreenAtDisplay (d, xid);
     if (s)
     {
-	SCALE_DISPLAY (s->display);
 	SCALE_SCREEN (s);
 
-	if (ss->state != SCALE_STATE_WAIT &&
-	    ss->state != SCALE_STATE_OUT)
+	if (ss->state != SCALE_STATE_WAIT && ss->state != SCALE_STATE_OUT)
 	{
-	    if (!layoutThumbs (s))
-		return FALSE;
-
-	    if (!ss->grabIndex)
-	    {
-		if (otherScreenGrabExist (s, "scale", 0))
-		    return FALSE;
-
-		ss->grabIndex = pushScreenGrab (s, ss->cursor, "scale");
-	    }
-
-	    if (ss->grabIndex)
-	    {
-		if (!sd->lastActiveNum)
-		    sd->lastActiveNum = s->activeNum - 1;
-
-		ss->state = SCALE_STATE_OUT;
-
-		damageScreen (s);
-	    }
-
-	    if (state & CompActionStateInitButton)
-		action->state |= CompActionStateTermButton;
-
-	    if (state & CompActionStateInitKey)
-		action->state |= CompActionStateTermKey;
+	    ss->type = ScaleTypeNormal;
+	    return scaleInitiateCommon (s, action, state, option, nOption);
 	}
     }
 
     return FALSE;
 }
 
+static Bool
+scaleInitiateAll (CompDisplay     *d,
+		  CompAction      *action,
+		  CompActionState state,
+		  CompOption      *option,
+		  int		  nOption)
+{
+    CompScreen *s;
+    Window     xid;
+
+    xid = getIntOptionNamed (option, nOption, "root", 0);
+
+    s = findScreenAtDisplay (d, xid);
+    if (s)
+    {
+	SCALE_SCREEN (s);
+
+	if (ss->state != SCALE_STATE_WAIT && ss->state != SCALE_STATE_OUT)
+	{
+	    ss->type = ScaleTypeAll;
+	    return scaleInitiateCommon (s, action, state, option, nOption);
+	}
+    }
+
+    return FALSE;
+}
+
+static Bool
+scaleInitiateGroup (CompDisplay     *d,
+		    CompAction      *action,
+		    CompActionState state,
+		    CompOption      *option,
+		    int		    nOption)
+{
+    CompScreen *s;
+    Window     xid;
+
+    xid = getIntOptionNamed (option, nOption, "root", 0);
+
+    s = findScreenAtDisplay (d, xid);
+    if (s)
+    {
+	SCALE_SCREEN (s);
+
+	if (ss->state != SCALE_STATE_WAIT && ss->state != SCALE_STATE_OUT)
+	{
+	    CompWindow *w;
+
+	    w = findWindowAtDisplay (d, getIntOptionNamed (option, nOption,
+							   "window", 0));
+	    if (w)
+	    {
+		ss->type	 = ScaleTypeGroup;
+		ss->clientLeader = (w->clientLeader) ? w->clientLeader : w->id;
+
+		return scaleInitiateCommon (s, action, state, option, nOption);
+	    }
+	}
+    }
+
+    return FALSE;
+}
+
+static Bool
+scaleInitiateOutput (CompDisplay     *d,
+		     CompAction      *action,
+		     CompActionState state,
+		     CompOption      *option,
+		     int	     nOption)
+{
+    CompScreen *s;
+    Window     xid;
+
+    xid = getIntOptionNamed (option, nOption, "root", 0);
+
+    s = findScreenAtDisplay (d, xid);
+    if (s)
+    {
+	SCALE_SCREEN (s);
+
+	if (ss->state != SCALE_STATE_WAIT && ss->state != SCALE_STATE_OUT)
+	{
+	    ss->type = ScaleTypeOutput;
+	    return scaleInitiateCommon (s, action, state, option, nOption);
+	}
+    }
+
+    return FALSE;
+}
+
+static void
+scaleSelectWindow (CompWindow *w)
+
+{
+    SCALE_DISPLAY (w->screen->display);
+
+    if (sd->selectedWindow != w->id)
+    {
+	CompWindow *old, *new;
+
+	old = findWindowAtScreen (w->screen, sd->selectedWindow);
+	new = findWindowAtScreen (w->screen, w->id);
+
+	sd->selectedWindow = w->id;
+
+	if (old)
+	    addWindowDamage (old);
+
+	addWindowDamage (new);
+    }
+}
+
 static Bool
 scaleSelectWindowAt (CompScreen *s,
 		     int	 x,
-		     int	 y)
+		     int	 y,
+		     Bool	 moveInputFocus)
 
 {
     CompWindow *w;
@@ -1097,7 +1475,17 @@ scaleSelectWindowAt (CompScreen *s,
     w = scaleCheckForWindowAt (s, x, y);
     if (w && isScaleWin (w))
     {
-	activateWindow (w);
+	SCALE_DISPLAY (s->display);
+
+	scaleSelectWindow (w);
+
+	if (moveInputFocus)
+	{
+	    sd->lastActiveNum    = w->activeNum;
+	    sd->lastActiveWindow = w->id;
+
+	    moveInputFocusToWindow (w);
+	}
 
 	return TRUE;
     }
@@ -1112,54 +1500,77 @@ scaleMoveFocusWindow (CompScreen *s,
 
 {
     CompWindow *active;
+    CompWindow *focus = NULL;
 
     active = findWindowAtScreen (s, s->display->activeWindow);
     if (active)
     {
-	CompWindow *w, *focus = NULL;
-	ScaleSlot  *slot;
-	int	   x, y, cx, cy, d, min = MAXSHORT;
-
-	SCALE_SCREEN (s);
 	SCALE_WINDOW (active);
 
-	if (!sw->slot)
-	    return;
+	if (sw->slot)
+	{
+	    SCALE_SCREEN (s);
 
-	cx = (sw->slot->x1 + sw->slot->x2) / 2;
-	cy = (sw->slot->y1 + sw->slot->y2) / 2;
+	    CompWindow *w;
+	    ScaleSlot  *slot;
+	    int	       x, y, cx, cy, d, min = MAXSHORT;
+
+	    cx = (sw->slot->x1 + sw->slot->x2) / 2;
+	    cy = (sw->slot->y1 + sw->slot->y2) / 2;
+
+	    for (w = s->windows; w; w = w->next)
+	    {
+		slot = GET_SCALE_WINDOW (w, ss)->slot;
+		if (!slot)
+		    continue;
+
+		x = (slot->x1 + slot->x2) / 2;
+		y = (slot->y1 + slot->y2) / 2;
+
+		d = abs (x - cx) + abs (y - cy);
+		if (d < min)
+		{
+		    if ((dx > 0 && slot->x1 < sw->slot->x2) ||
+			(dx < 0 && slot->x2 > sw->slot->x1) ||
+			(dy > 0 && slot->y1 < sw->slot->y2) ||
+			(dy < 0 && slot->y2 > sw->slot->y1))
+			continue;
+
+		    min   = d;
+		    focus = w;
+		}
+	    }
+	}
+    }
+
+    /* move focus to the last focused window if no slot window is currently
+       focused */
+    if (!focus)
+    {
+	CompWindow *w;
+
+	SCALE_SCREEN (s);
 
 	for (w = s->windows; w; w = w->next)
 	{
-	    slot = GET_SCALE_WINDOW (w, ss)->slot;
-	    if (!slot)
+	    if (!GET_SCALE_WINDOW (w, ss)->slot)
 		continue;
 
-	    x = (slot->x1 + slot->x2) / 2;
-	    y = (slot->y1 + slot->y2) / 2;
-
-	    d = abs (x - cx) + abs (y - cy);
-	    if (d < min)
-	    {
-		if ((dx > 0 && slot->x1 < sw->slot->x2) ||
-		    (dx < 0 && slot->x2 > sw->slot->x1) ||
-		    (dy > 0 && slot->y1 < sw->slot->y2) ||
-		    (dy < 0 && slot->y2 > sw->slot->y1))
-		    continue;
-
-		min   = d;
+	    if (!focus || focus->activeNum < w->activeNum)
 		focus = w;
-	    }
 	}
+    }
 
-	if (focus)
-	{
-	    SCALE_DISPLAY (s->display);
+    if (focus)
+    {
+	SCALE_DISPLAY (s->display);
 
-	    sd->lastActiveNum = focus->activeNum;
+	scaleSelectWindow (focus);
 
-	    activateWindow (focus);
-	}
+	sd->lastActiveNum    = focus->activeNum;
+	sd->lastActiveWindow = focus->id;
+
+	moveInputFocusToWindow (focus);
     }
 }
 
@@ -1174,7 +1585,7 @@ scaleWindowRemove (CompDisplay *d,
     {
 	SCALE_SCREEN (w->screen);
 
-	if (ss->grabIndex && ss->state != SCALE_STATE_IN)
+	if (ss->state != SCALE_STATE_NONE && ss->state != SCALE_STATE_IN)
 	{
 	    int i;
 
@@ -1194,6 +1605,42 @@ scaleWindowRemove (CompDisplay *d,
     }
 }
 
+static Bool
+scaleHoverTimeout (void *closure)
+{
+    CompScreen *s = closure;
+
+    SCALE_SCREEN (s);
+    SCALE_DISPLAY (s->display);
+
+    if (ss->grab && ss->state != SCALE_STATE_IN)
+    {
+	CompWindow *w;
+	CompOption o;
+	CompAction *action =
+	    &sd->opt[SCALE_DISPLAY_OPTION_INITIATE].value.action;
+
+	w = findWindowAtDisplay (s->display, sd->selectedWindow);
+	if (w)
+	{
+	    sd->lastActiveNum    = w->activeNum;
+	    sd->lastActiveWindow = w->id;
+
+	    moveInputFocusToWindow (w);
+	}
+
+	o.type    = CompOptionTypeInt;
+	o.name    = "root";
+	o.value.i = s->root;
+
+	scaleTerminate (s->display, action, 0, &o, 1);
+    }
+
+    ss->hoverHandle = 0;
+
+    return FALSE;
+}
+
 static void
 scaleHandleEvent (CompDisplay *d,
 		  XEvent      *event)
@@ -1243,7 +1690,8 @@ scaleHandleEvent (CompDisplay *d,
 
 		    if (scaleSelectWindowAt (s,
 					     event->xbutton.x_root,
-					     event->xbutton.y_root))
+					     event->xbutton.y_root,
+					     TRUE))
 		    {
 			scaleTerminate (d, action, 0, &o, 1);
 		    }
@@ -1267,12 +1715,105 @@ scaleHandleEvent (CompDisplay *d,
 	{
 	    SCALE_SCREEN (s);
 
-	    if (ss->grabIndex		    &&
-		ss->state != SCALE_STATE_IN &&
-		ss->opt[SCALE_SCREEN_OPTION_SLOPPY_FOCUS].value.b)
+	    if (ss->grabIndex && ss->state != SCALE_STATE_IN)
+	    {
+		Bool focus;
+
+		focus = !d->opt[COMP_DISPLAY_OPTION_CLICK_TO_FOCUS].value.b;
+
 		scaleSelectWindowAt (s,
 				     event->xmotion.x_root,
-				     event->xmotion.y_root);
+				     event->xmotion.y_root,
+				     focus);
+	    }
+	}
+	break;
+    case ClientMessage:
+	if (event->xclient.message_type == d->xdndPositionAtom)
+	{
+	    CompWindow *w;
+
+	    w = findWindowAtDisplay (d, event->xclient.window);
+	    if (w)
+	    {
+		Bool focus;
+
+		SCALE_SCREEN (w->screen);
+
+		s = w->screen;
+
+		focus = !d->opt[COMP_DISPLAY_OPTION_CLICK_TO_FOCUS].value.b;
+
+		if (w->id == ss->dndTarget)
+		    sendDndStatusMessage (w->screen, event->xclient.data.l[0]);
+
+		if (ss->grab			&&
+		    ss->state != SCALE_STATE_IN &&
+		    w->id == ss->dndTarget)
+		{
+		    int x = event->xclient.data.l[2] >> 16;
+		    int y = event->xclient.data.l[2] & 0xffff;
+
+		    w = scaleCheckForWindowAt (s, x, y);
+		    if (w && isScaleWin (w))
+		    {
+			int time;
+
+			time = ss->opt[SCALE_SCREEN_OPTION_HOVER_TIME].value.i;
+
+			if (ss->hoverHandle)
+			{
+			    if (w->id != sd->selectedWindow)
+			    {
+				compRemoveTimeout (ss->hoverHandle);
+				ss->hoverHandle = 0;
+			    }
+			}
+
+			if (!ss->hoverHandle)
+			    ss->hoverHandle =
+				compAddTimeout (time,
+						scaleHoverTimeout,
+						s);
+
+			scaleSelectWindowAt (s, x, y, focus);
+		    }
+		    else
+		    {
+			if (ss->hoverHandle)
+			    compRemoveTimeout (ss->hoverHandle);
+
+			ss->hoverHandle = 0;
+		    }
+		}
+	    }
+	}
+	else if (event->xclient.message_type == d->xdndDropAtom ||
+		 event->xclient.message_type == d->xdndLeaveAtom)
+	{
+	    CompWindow *w;
+
+	    w = findWindowAtDisplay (d, event->xclient.window);
+	    if (w)
+	    {
+		CompAction *action =
+		    &sd->opt[SCALE_DISPLAY_OPTION_INITIATE].value.action;
+
+		SCALE_SCREEN (w->screen);
+
+		if (ss->grab			&&
+		    ss->state != SCALE_STATE_IN &&
+		    w->id == ss->dndTarget)
+		{
+		    CompOption o;
+
+		    o.type    = CompOptionTypeInt;
+		    o.name    = "root";
+		    o.value.i = w->screen->root;
+
+		    scaleTerminate (d, action, 0, &o, 1);
+		}
+	    }
 	}
     default:
 	break;
@@ -1298,24 +1839,40 @@ scaleDamageWindowRect (CompWindow *w,
 		       Bool	  initial,
 		       BoxPtr     rect)
 {
-    Bool status;
+    Bool status = FALSE;
 
     SCALE_SCREEN (w->screen);
 
     if (initial)
     {
-	if (isScaleWin (w))
+	if (ss->grab && isScaleWin (w))
 	{
-	    if (ss->grabIndex && layoutThumbs (w->screen))
+	    if (layoutThumbs (w->screen))
 	    {
 		ss->state = SCALE_STATE_OUT;
 		damageScreen (w->screen);
 	    }
 	}
     }
+    else if (ss->state == SCALE_STATE_WAIT)
+    {
+	SCALE_WINDOW (w);
+
+	if (sw->slot)
+	{
+	    damageTransformedWindowRect (w,
+					 sw->scale,
+					 sw->scale,
+					 sw->tx,
+					 sw->ty,
+					 rect);
+
+	    status = TRUE;
+	}
+    }
 
     UNWRAP (ss, w->screen, damageWindowRect);
-    status = (*w->screen->damageWindowRect) (w, initial, rect);
+    status |= (*w->screen->damageWindowRect) (w, initial, rect);
     WRAP (ss, w->screen, damageWindowRect, scaleDamageWindowRect);
 
     return status;
@@ -1348,6 +1905,9 @@ scaleSetDisplayOption (CompDisplay     *display,
 
     switch (index) {
     case SCALE_DISPLAY_OPTION_INITIATE:
+    case SCALE_DISPLAY_OPTION_INITIATE_ALL:
+    case SCALE_DISPLAY_OPTION_INITIATE_GROUP:
+    case SCALE_DISPLAY_OPTION_INITIATE_OUTPUT:
 	if (setDisplayAction (display, o, value))
 	    return TRUE;
     default:
@@ -1373,12 +1933,62 @@ scaleDisplayInitOptions (ScaleDisplay *sd,
     o->value.action.bell	  = FALSE;
     o->value.action.edgeMask	  = (1 << SCREEN_EDGE_TOPRIGHT);
     o->value.action.state	  = CompActionStateInitEdge;
+    o->value.action.state	 |= CompActionStateInitEdgeDnd;
     o->value.action.type	  = CompBindingTypeKey;
     o->value.action.state	 |= CompActionStateInitKey;
+    o->value.action.state	 |= CompActionStateInitButton;
     o->value.action.key.modifiers = SCALE_INITIATE_MODIFIERS_DEFAULT;
     o->value.action.key.keycode   =
 	XKeysymToKeycode (display,
 			  XStringToKeysym (SCALE_INITIATE_KEY_DEFAULT));
+
+    o = &sd->opt[SCALE_DISPLAY_OPTION_INITIATE_ALL];
+    o->name		      = "initiate_all";
+    o->shortDesc	      = N_("Initiate Window Picker For All Windows");
+    o->longDesc		      = N_("Layout and start transforming all windows");
+    o->type		      = CompOptionTypeAction;
+    o->value.action.initiate  = scaleInitiateAll;
+    o->value.action.terminate = scaleTerminate;
+    o->value.action.bell      = FALSE;
+    o->value.action.edgeMask  = 0;
+    o->value.action.state     = CompActionStateInitEdge;
+    o->value.action.state    |= CompActionStateInitEdgeDnd;
+    o->value.action.type      = 0;
+    o->value.action.state    |= CompActionStateInitKey;
+    o->value.action.state    |= CompActionStateInitButton;
+
+    o = &sd->opt[SCALE_DISPLAY_OPTION_INITIATE_GROUP];
+    o->name		      = "initiate_group";
+    o->shortDesc	      = N_("Initiate Window Picker For Window Group");
+    o->longDesc		      = N_("Layout and start transforming "
+				   "window group");
+    o->type		      = CompOptionTypeAction;
+    o->value.action.initiate  = scaleInitiateGroup;
+    o->value.action.terminate = scaleTerminate;
+    o->value.action.bell      = FALSE;
+    o->value.action.edgeMask  = 0;
+    o->value.action.state     = CompActionStateInitEdge;
+    o->value.action.state    |= CompActionStateInitEdgeDnd;
+    o->value.action.type      = 0;
+    o->value.action.state    |= CompActionStateInitKey;
+    o->value.action.state    |= CompActionStateInitButton;
+
+    o = &sd->opt[SCALE_DISPLAY_OPTION_INITIATE_OUTPUT];
+    o->name		      = "initiate_output";
+    o->shortDesc	      = N_("Initiate Window Picker For Windows on "
+				   "Current Output");
+    o->longDesc		      = N_("Layout and start transforming "
+				   "windows on current output");
+    o->type		      = CompOptionTypeAction;
+    o->value.action.initiate  = scaleInitiateOutput;
+    o->value.action.terminate = scaleTerminate;
+    o->value.action.bell      = FALSE;
+    o->value.action.edgeMask  = 0;
+    o->value.action.state     = CompActionStateInitEdge;
+    o->value.action.state    |= CompActionStateInitEdgeDnd;
+    o->value.action.type      = 0;
+    o->value.action.state    |= CompActionStateInitKey;
+    o->value.action.state    |= CompActionStateInitButton;
 }
 
 static Bool
@@ -1399,6 +2009,7 @@ scaleInitDisplay (CompPlugin  *p,
     }
 
     sd->lastActiveNum = None;
+    sd->selectedWindow = None;
 
     scaleDisplayInitOptions (sd, d->display);
 
@@ -1446,8 +2057,13 @@ scaleInitScreen (CompPlugin *p,
 	return FALSE;
     }
 
+    ss->grab      = FALSE;
     ss->grabIndex = 0;
 
+    ss->hoverHandle = 0;
+
+    ss->dndTarget = None;
+
     ss->state = SCALE_STATE_NONE;
 
     ss->slots = 0;
@@ -1469,6 +2085,12 @@ scaleInitScreen (CompPlugin *p,
     scaleScreenInitOptions (ss);
 
     addScreenAction (s, &sd->opt[SCALE_DISPLAY_OPTION_INITIATE].value.action);
+    addScreenAction (s,
+		     &sd->opt[SCALE_DISPLAY_OPTION_INITIATE_ALL].value.action);
+    addScreenAction (s,
+		     &sd->opt[SCALE_DISPLAY_OPTION_INITIATE_GROUP].value.action);
+    addScreenAction (s,
+		     &sd->opt[SCALE_DISPLAY_OPTION_INITIATE_OUTPUT].value.action);
 
     WRAP (ss, s, preparePaintScreen, scalePreparePaintScreen);
     WRAP (ss, s, donePaintScreen, scaleDonePaintScreen);
@@ -1523,6 +2145,7 @@ scaleInitWindow (CompPlugin *p,
     sw->xVelocity = sw->yVelocity = 0.0f;
     sw->scaleVelocity = 1.0f;
     sw->delta = 1.0f;
+    sw->lastThumbOpacity = 0.0f;
 
     w->privates[ss->windowPrivateIndex].ptr = sw;
 
