@@ -46,7 +46,7 @@ handleWindowDamageRect (CompWindow *w,
     REGION region;
     Bool   initial = FALSE;
 
-    if (!w->redirected)
+    if (!w->redirected || w->bindFailed)
 	return;
 
     if (!w->damaged)
@@ -91,6 +91,8 @@ handleSyncAlarm (CompWindow *w)
 	    w->syncWaitHandle = 0;
 	}
 
+	w->syncWait = FALSE;
+
 	if (resizeWindow (w,
 			  w->syncX, w->syncY,
 			  w->syncWidth, w->syncHeight,
@@ -111,7 +113,12 @@ handleSyncAlarm (CompWindow *w)
 	    }
 
 	    w->nDamage = 0;
-	    w->syncWait = FALSE;
+	}
+	else
+	{
+	    /* resizeWindow failing means that there is another pending
+	       resize and we must send a new sync request to the client */
+	    sendSyncRequest (w);
 	}
     }
 }
@@ -1342,25 +1349,7 @@ handleEvent (CompDisplay *d,
 	}
 	break;
     case PropertyNotify:
-	if (event->xproperty.atom == d->winActiveAtom)
-	{
-	    Window newActiveWindow;
-
-	    newActiveWindow = getActiveWindow (d, event->xproperty.window);
-	    if (newActiveWindow != d->activeWindow)
-	    {
-		d->activeWindow = newActiveWindow;
-
-		s = findScreenAtDisplay (d, event->xproperty.window);
-		if (s)
-		{
-		    w = findWindowAtDisplay (d, newActiveWindow);
-		    if (w)
-			w->activeNum = s->activeNum++;
-		}
-	    }
-	}
-	else if (event->xproperty.atom == d->winTypeAtom)
+	if (event->xproperty.atom == d->winTypeAtom)
 	{
 	    w = findWindowAtDisplay (d, event->xproperty.window);
 	    if (w)
@@ -1879,6 +1868,8 @@ handleEvent (CompDisplay *d,
 		w->initialViewportX = w->screen->x;
 		w->initialViewportY = w->screen->y;
 
+		w->initialTimestampSet = FALSE;
+
 		applyStartupProperties (w->screen, w);
 
 		w->pendingMaps++;
@@ -1888,7 +1879,21 @@ handleEvent (CompDisplay *d,
 		updateWindowAttributes (w, FALSE);
 
 		if (focusWindowOnMap (w))
+		{
 		    moveInputFocusToWindow (w);
+		}
+		else if (w->type & ~CompWindowTypeSplashMask)
+		{
+		    CompWindow *p;
+
+		    for (p = w->prev; p; p = p->prev)
+			if (p->id == d->activeWindow)
+			    break;
+
+		    /* window is above active window so we should lower it */
+		    if (p)
+			restackWindowBelow (w, p);
+		}
 	    }
 
 	    setWindowProp (d, w->id, d->winDesktopAtom, w->desktop);
@@ -1940,16 +1945,21 @@ handleEvent (CompDisplay *d,
     case FocusIn:
 	if (event->xfocus.mode != NotifyGrab)
 	{
-	    w = findWindowAtDisplay (d, event->xfocus.window);
+	    w = findTopLevelWindowAtDisplay (d, event->xfocus.window);
 	    if (w && w->managed)
 	    {
 		unsigned int state = w->state;
 
 		if (w->id != d->activeWindow)
+		{
+		    d->activeWindow = w->id;
+		    w->activeNum = w->screen->activeNum++;
+
 		    XChangeProperty (d->display, w->screen->root,
 				     d->winActiveAtom,
 				     XA_WINDOW, 32, PropModeReplace,
 				     (unsigned char *) &w->id, 1);
+		}
 
 		w->state &= ~CompWindowStateDemandsAttentionMask;
 
@@ -1961,6 +1971,7 @@ handleEvent (CompDisplay *d,
     case EnterNotify:
 	if (!d->screens->maxGrab		    &&
 	    event->xcrossing.mode   != NotifyGrab   &&
+	    event->xcrossing.mode   != NotifyUngrab &&
 	    event->xcrossing.detail != NotifyInferior)
 	{
 	    Bool raise;
