@@ -47,18 +47,39 @@ donePaintScreen (CompScreen *screen) {}
 void
 applyScreenTransform (CompScreen	      *screen,
 		      const ScreenPaintAttrib *sAttrib,
-		      int		      output)
+		      int		      output,
+		      CompTransform	      *transform)
 {
-    glTranslatef (sAttrib->xTranslate,
-		  sAttrib->yTranslate,
-		  sAttrib->zTranslate + sAttrib->zCamera);
+    matrixTranslate (transform,
+		     sAttrib->xTranslate,
+		     sAttrib->yTranslate,
+		     sAttrib->zTranslate + sAttrib->zCamera);
+    matrixRotate (transform,
+		  sAttrib->xRotate, 0.0f, 1.0f, 0.0f);
+    matrixRotate (transform,
+		  sAttrib->vRotate,
+		  1.0f - sAttrib->xRotate / 90.0f,
+		  0.0f,
+		  sAttrib->xRotate / 90.0f);
+    matrixRotate (transform,
+		  sAttrib->yRotate, 0.0f, 1.0f, 0.0f);
+}
 
-    glRotatef (sAttrib->xRotate, 0.0f, 1.0f, 0.0f);
-    glRotatef (sAttrib->vRotate,
-	       1.0f - sAttrib->xRotate / 90.0f,
-	       0.0f,
-	       sAttrib->xRotate / 90.0f);
-    glRotatef (sAttrib->yRotate, 0.0f, 1.0f, 0.0f);
+void
+transformToScreenSpace (CompScreen    *screen,
+			int	      output,
+			float         z,
+			CompTransform *transform)
+{
+    matrixTranslate (transform, -0.5f, -0.5f, z);
+    matrixScale (transform,
+		 1.0f  / screen->outputDev[output].width,
+		 -1.0f / screen->outputDev[output].height,
+		 1.0f);
+    matrixTranslate (transform,
+		     -screen->outputDev[output].region.extents.x1,
+		     -screen->outputDev[output].region.extents.y2,
+		     0.0f);
 }
 
 void
@@ -78,24 +99,22 @@ prepareXCoords (CompScreen *screen,
 void
 paintTransformedScreen (CompScreen		*screen,
 			const ScreenPaintAttrib *sAttrib,
+			const CompTransform	*transform,
 			Region			region,
 			int			output,
 			unsigned int		mask)
 {
-    CompWindow *w;
-    int	       windowMask;
-    int	       backgroundMask;
+    CompTransform sTransform = *transform;
+    CompWindow	  *w;
+    int	          windowMask;
+    int	          backgroundMask;
 
     if (mask & PAINT_SCREEN_CLEAR_MASK)
 	clearTargetOutput (screen->display, GL_COLOR_BUFFER_BIT);
 
     screenLighting (screen, TRUE);
 
-    glPushMatrix ();
-
-    (screen->applyScreenTransform) (screen, sAttrib, output);
-
-    prepareXCoords (screen, output, -sAttrib->zTranslate);
+    (*screen->applyScreenTransform) (screen, sAttrib, output, &sTransform);
 
     if (mask & PAINT_SCREEN_TRANSFORMED_MASK)
     {
@@ -104,12 +123,43 @@ paintTransformedScreen (CompScreen		*screen,
 
 	if (mask & PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK)
 	{
-	    backgroundMask |= PAINT_BACKGROUND_WITH_STENCIL_MASK;
+	    GLdouble h = screen->height;
+
+	    GLdouble p1[2] = { region->extents.x1, h - region->extents.y2 };
+	    GLdouble p2[2] = { region->extents.x2, h - region->extents.y1 };
+
+	    GLdouble halfW = screen->outputDev[output].width / 2.0;
+	    GLdouble halfH = screen->outputDev[output].height / 2.0;
+
+	    GLdouble cx = screen->outputDev[output].region.extents.x1 + halfW;
+	    GLdouble cy = (h - screen->outputDev[output].region.extents.y2) +
+		halfH;
+
+	    GLdouble top[4]    = { 0.0, halfH / (cy - p1[1]), 0.0, 0.5 };
+	    GLdouble bottom[4] = { 0.0, halfH / (cy - p2[1]), 0.0, 0.5 };
+	    GLdouble left[4]   = { halfW / (cx - p1[0]), 0.0, 0.0, 0.5 };
+	    GLdouble right[4]  = { halfW / (cx - p2[0]), 0.0, 0.0, 0.5 };
+
+	    glPushMatrix ();
+	    glLoadMatrixf (sTransform.m);
+
+	    glClipPlane (GL_CLIP_PLANE0, top);
+	    glClipPlane (GL_CLIP_PLANE1, bottom);
+	    glClipPlane (GL_CLIP_PLANE2, left);
+	    glClipPlane (GL_CLIP_PLANE3, right);
+
+	    glEnable (GL_CLIP_PLANE0);
+	    glEnable (GL_CLIP_PLANE1);
+	    glEnable (GL_CLIP_PLANE2);
+	    glEnable (GL_CLIP_PLANE3);
+
+	    transformToScreenSpace (screen, output, -sAttrib->zTranslate,
+				    &sTransform);
+
+	    glLoadMatrixf (sTransform.m);
 
 	    (*screen->paintBackground) (screen, region, backgroundMask);
 
-	    glEnable (GL_STENCIL_TEST);
-
 	    for (w = screen->windows; w; w = w->next)
 	    {
 		if (w->destroyed)
@@ -121,10 +171,14 @@ paintTransformedScreen (CompScreen		*screen,
 			continue;
 		}
 
-		(*screen->paintWindow) (w, &w->paint, region, windowMask);
+		(*screen->paintWindow) (w, &w->paint, &sTransform, region,
+					windowMask);
 	    }
 
-	    glDisable (GL_STENCIL_TEST);
+	    glDisable (GL_CLIP_PLANE0);
+	    glDisable (GL_CLIP_PLANE1);
+	    glDisable (GL_CLIP_PLANE2);
+	    glDisable (GL_CLIP_PLANE3);
 
 	    glPopMatrix ();
 
@@ -134,6 +188,11 @@ paintTransformedScreen (CompScreen		*screen,
     else
 	windowMask = backgroundMask = 0;
 
+    transformToScreenSpace (screen, output, -sAttrib->zTranslate, &sTransform);
+
+    glPushMatrix ();
+    glLoadMatrixf (sTransform.m);
+
     (*screen->paintBackground) (screen, region, backgroundMask);
 
     for (w = screen->windows; w; w = w->next)
@@ -147,7 +206,7 @@ paintTransformedScreen (CompScreen		*screen,
 		continue;
 	}
 
-	(*screen->paintWindow) (w, &w->paint, region, windowMask);
+	(*screen->paintWindow) (w, &w->paint, &sTransform, region, windowMask);
     }
 
     glPopMatrix ();
@@ -156,11 +215,13 @@ paintTransformedScreen (CompScreen		*screen,
 Bool
 paintScreen (CompScreen		     *screen,
 	     const ScreenPaintAttrib *sAttrib,
+	     const CompTransform     *transform,
 	     Region		     region,
 	     int		     output,
 	     unsigned int	     mask)
 {
     static Region tmpRegion = NULL;
+    CompTransform sTransform = *transform;
     CompWindow	  *w;
 
     if (mask & PAINT_SCREEN_REGION_MASK)
@@ -171,7 +232,8 @@ paintScreen (CompScreen		     *screen,
 	    {
 		region = &screen->outputDev[output].region;
 
-		(*screen->paintTransformedScreen) (screen, sAttrib, region,
+		(*screen->paintTransformedScreen) (screen, sAttrib,
+						   &sTransform, region,
 						   output, mask);
 
 		return TRUE;
@@ -184,7 +246,7 @@ paintScreen (CompScreen		     *screen,
     }
     else if (mask & PAINT_SCREEN_FULL_MASK)
     {
-	(*screen->paintTransformedScreen) (screen, sAttrib,
+	(*screen->paintTransformedScreen) (screen, sAttrib, &sTransform,
 					   &screen->outputDev[output].region,
 					   output, mask);
 
@@ -195,9 +257,10 @@ paintScreen (CompScreen		     *screen,
 
     screenLighting (screen, FALSE);
 
-    glPushMatrix ();
+    transformToScreenSpace (screen, output, -DEFAULT_Z_CAMERA, &sTransform);
 
-    prepareXCoords (screen, output, -DEFAULT_Z_CAMERA);
+    glPushMatrix ();
+    glLoadMatrixf (sTransform.m);
 
     if (mask & PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK)
     {
@@ -215,7 +278,7 @@ paintScreen (CompScreen		     *screen,
 		    continue;
 	    }
 
-	    (*screen->paintWindow) (w, &w->paint, region, 0);
+	    (*screen->paintWindow) (w, &w->paint, &sTransform, region, 0);
 	}
     }
     else
@@ -243,8 +306,8 @@ paintScreen (CompScreen		     *screen,
 		    continue;
 	    }
 
-	    if ((*screen->paintWindow) (w, &w->paint, tmpRegion,
-					PAINT_WINDOW_SOLID_MASK))
+	    if ((*screen->paintWindow) (w, &w->paint, &sTransform, tmpRegion,
+					PAINT_WINDOW_CLIP_OPAQUE_MASK))
 	    {
 		XSubtractRegion (tmpRegion, w->region, tmpRegion);
 
@@ -279,8 +342,8 @@ paintScreen (CompScreen		     *screen,
 		    continue;
 	    }
 
-	    (*screen->paintWindow) (w, &w->paint, w->clip,
-				    PAINT_WINDOW_TRANSLUCENT_MASK);
+	    (*screen->paintWindow) (w, &w->paint, &sTransform, w->clip,
+				    PAINT_WINDOW_CLIP_TRANSLUCENT_MASK);
 	}
     }
 
@@ -388,6 +451,46 @@ moreWindowIndices (CompWindow *w,
     return TRUE;
 }
 
+static void
+drawWindowGeometry (CompWindow *w)
+{
+    int     texUnit = w->texUnits;
+    int     currentTexUnit = 0;
+    int     stride = (1 + texUnit) * 2;
+    GLfloat *vertices = w->vertices + (stride - 2);
+
+    stride *= sizeof (GLfloat);
+
+    glVertexPointer (2, GL_FLOAT, stride, vertices);
+
+    while (texUnit--)
+    {
+	if (texUnit != currentTexUnit)
+	{
+	    (*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB + texUnit);
+	    glEnableClientState (GL_TEXTURE_COORD_ARRAY);
+	    currentTexUnit = texUnit;
+	}
+	vertices -= 2;
+	glTexCoordPointer (2, GL_FLOAT, stride, vertices);
+    }
+
+    glDrawArrays (GL_QUADS, 0, w->vCount);
+
+    /* disable all texture coordinate arrays except 0 */
+    texUnit = w->texUnits;
+    if (texUnit > 1)
+    {
+	while (--texUnit)
+	{
+	    (*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB + texUnit);
+	    glDisableClientState (GL_TEXTURE_COORD_ARRAY);
+	}
+
+	(*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB);
+    }
+}
+
 void
 addWindowGeometry (CompWindow *w,
 		   CompMatrix *matrix,
@@ -526,87 +629,111 @@ addWindowGeometry (CompWindow *w,
 		}
 	    }
 	}
-	w->vCount = n * 4;
+
+	w->vCount	      = n * 4;
+	w->drawWindowGeometry = drawWindowGeometry;
     }
 }
 
-void
-drawWindowGeometry (CompWindow *w)
+static Bool
+enableFragmentProgramAndDrawGeometry (CompWindow	   *w,
+				      CompTexture	   *texture,
+				      const FragmentAttrib *attrib,
+				      int		   filter,
+				      unsigned int	   mask)
 {
-    int     texUnit = w->texUnits;
-    int     currentTexUnit = 0;
-    int     stride = (1 + texUnit) * 2;
-    GLfloat *vertices = w->vertices + (stride - 2);
+    FragmentAttrib fa = *attrib;
+    CompScreen     *s = w->screen;
+    Bool	   blending;
 
-    stride *= sizeof (GLfloat);
-
-    glVertexPointer (2, GL_FLOAT, stride, vertices);
-
-    while (texUnit--)
+    if (s->canDoSaturated && attrib->saturation != COLOR)
     {
-	if (texUnit != currentTexUnit)
-	{
-	    (*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB + texUnit);
-	    glEnableClientState (GL_TEXTURE_COORD_ARRAY);
-	    currentTexUnit = texUnit;
-	}
-	vertices -= 2;
-	glTexCoordPointer (2, GL_FLOAT, stride, vertices);
+	int param, function;
+
+	param    = allocFragmentParameters (&fa, 1);
+	function = getSaturateFragmentFunction (s, texture, param);
+
+	addFragmentFunction (&fa, function);
+
+	(*s->programEnvParameter4f) (GL_FRAGMENT_PROGRAM_ARB, param,
+				     RED_SATURATION_WEIGHT,
+				     GREEN_SATURATION_WEIGHT,
+				     BLUE_SATURATION_WEIGHT,
+				     attrib->saturation / 65535.0f);
     }
 
-    glDrawArrays (GL_QUADS, 0, w->vCount);
+    if (!enableFragmentAttrib (s, &fa, &blending))
+	return FALSE;
 
-    /* disable all texture coordinate arrays except 0 */
-    texUnit = w->texUnits;
-    if (texUnit > 1)
+    enableTexture (s, texture, filter);
+
+    if (mask & PAINT_WINDOW_BLEND_MASK)
     {
-	while (--texUnit)
+	if (blending)
+	    glEnable (GL_BLEND);
+
+	if (attrib->opacity != OPAQUE || attrib->brightness != BRIGHT)
 	{
-	    (*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB + texUnit);
-	    glDisableClientState (GL_TEXTURE_COORD_ARRAY);
+	    GLushort color;
+
+	    color = (attrib->opacity * attrib->brightness) >> 16;
+
+	    screenTexEnvMode (s, GL_MODULATE);
+	    glColor4us (color, color, color, attrib->opacity);
+
+	    (*w->drawWindowGeometry) (w);
+
+	    glColor4usv (defaultColor);
+	    screenTexEnvMode (s, GL_REPLACE);
+	}
+	else
+	{
+	    (*w->drawWindowGeometry) (w);
 	}
 
-	(*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB);
+	if (blending)
+	    glDisable (GL_BLEND);
     }
-}
-
-void
-drawWindowTexture (CompWindow		   *w,
-		   CompTexture		   *texture,
-		   const WindowPaintAttrib *attrib,
-		   unsigned int		   mask)
-{
-    int filter;
-
-    glPushMatrix ();
-
-    if (mask & PAINT_WINDOW_TRANSFORMED_MASK)
+    else if (attrib->brightness != BRIGHT)
     {
-	glTranslatef (w->attrib.x, w->attrib.y, 0.0f);
-	glScalef (attrib->xScale, attrib->yScale, 0.0f);
-	glTranslatef (attrib->xTranslate / attrib->xScale - w->attrib.x,
-		      attrib->yTranslate / attrib->yScale - w->attrib.y,
-		      0.0f);
+	screenTexEnvMode (s, GL_MODULATE);
+	glColor4us (attrib->brightness, attrib->brightness,
+		    attrib->brightness, BRIGHT);
 
-	filter = w->screen->filter[WINDOW_TRANS_FILTER];
-    }
-    else if (mask & PAINT_WINDOW_ON_TRANSFORMED_SCREEN_MASK)
-    {
-	filter = w->screen->filter[SCREEN_TRANS_FILTER];
+	(*w->drawWindowGeometry) (w);
+
+	glColor4usv (defaultColor);
+	screenTexEnvMode (s, GL_REPLACE);
     }
     else
     {
-	filter = w->screen->filter[NOTHING_TRANS_FILTER];
+	(*w->drawWindowGeometry) (w);
     }
 
-    if (w->screen->canDoSaturated && attrib->saturation != COLOR)
+    disableTexture (w->screen, texture);
+
+    disableFragmentAttrib (s, &fa);
+
+    return TRUE;
+}
+
+static void
+enableFragmentOperationsAndDrawGeometry (CompWindow	      *w,
+					 CompTexture	      *texture,
+					 const FragmentAttrib *attrib,
+					 int		      filter,
+					 unsigned int	      mask)
+{
+    CompScreen *s = w->screen;
+
+    if (s->canDoSaturated && attrib->saturation != COLOR)
     {
 	GLfloat constant[4];
 
-	if (mask & PAINT_WINDOW_TRANSLUCENT_MASK)
+	if (mask & PAINT_WINDOW_BLEND_MASK)
 	    glEnable (GL_BLEND);
 
-	enableTexture (w->screen, texture, filter);
+	enableTexture (s, texture, filter);
 
 	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
 
@@ -624,9 +751,9 @@ drawWindowTexture (CompWindow		   *w,
 
 	glColor4f (1.0f, 1.0f, 1.0f, 0.5f);
 
-	w->screen->activeTexture (GL_TEXTURE1_ARB);
+	s->activeTexture (GL_TEXTURE1_ARB);
 
-	enableTexture (w->screen, texture, filter);
+	enableTexture (s, texture, filter);
 
 	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
 
@@ -636,7 +763,7 @@ drawWindowTexture (CompWindow		   *w,
 	glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
 	glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
 
-	if (w->screen->canDoSlightlySaturated && attrib->saturation > 0)
+	if (s->canDoSlightlySaturated && attrib->saturation > 0)
 	{
 	    glTexEnvf (GL_TEXTURE_ENV, GL_COMBINE_ALPHA, GL_REPLACE);
 	    glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA, GL_PREVIOUS);
@@ -649,9 +776,9 @@ drawWindowTexture (CompWindow		   *w,
 
 	    glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, constant);
 
-	    w->screen->activeTexture (GL_TEXTURE2_ARB);
+	    s->activeTexture (GL_TEXTURE2_ARB);
 
-	    enableTexture (w->screen, texture, filter);
+	    enableTexture (s, texture, filter);
 
 	    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
 
@@ -673,9 +800,9 @@ drawWindowTexture (CompWindow		   *w,
 
 	    if (attrib->opacity < OPAQUE || attrib->brightness != BRIGHT)
 	    {
-		w->screen->activeTexture (GL_TEXTURE3_ARB);
+		s->activeTexture (GL_TEXTURE3_ARB);
 
-		enableTexture (w->screen, texture, filter);
+		enableTexture (s, texture, filter);
 
 		constant[3] = attrib->opacity / 65535.0f;
 		constant[0] = constant[1] = constant[2] = constant[3] *
@@ -697,24 +824,24 @@ drawWindowTexture (CompWindow		   *w,
 		glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA, GL_SRC_ALPHA);
 		glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND1_ALPHA, GL_SRC_ALPHA);
 
-		(*w->screen->drawWindowGeometry) (w);
+		(*w->drawWindowGeometry) (w);
 
-		disableTexture (w->screen, texture);
+		disableTexture (s, texture);
 
 		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
 
-		w->screen->activeTexture (GL_TEXTURE2_ARB);
+		s->activeTexture (GL_TEXTURE2_ARB);
 	    }
 	    else
 	    {
-		(*w->screen->drawWindowGeometry) (w);
+		(*w->drawWindowGeometry) (w);
 	    }
 
-	    disableTexture (w->screen, texture);
+	    disableTexture (s, texture);
 
 	    glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
 
-	    w->screen->activeTexture (GL_TEXTURE1_ARB);
+	    s->activeTexture (GL_TEXTURE1_ARB);
 	}
 	else
 	{
@@ -734,28 +861,28 @@ drawWindowTexture (CompWindow		   *w,
 
 	    glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, constant);
 
-	    (*w->screen->drawWindowGeometry) (w);
+	    (*w->drawWindowGeometry) (w);
 	}
 
-	disableTexture (w->screen, texture);
+	disableTexture (s, texture);
 
 	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
 
-	w->screen->activeTexture (GL_TEXTURE0_ARB);
+	s->activeTexture (GL_TEXTURE0_ARB);
 
-	disableTexture (w->screen, texture);
+	disableTexture (s, texture);
 
 	glColor4usv (defaultColor);
-	screenTexEnvMode (w->screen, GL_REPLACE);
+	screenTexEnvMode (s, GL_REPLACE);
 
-	if (mask & PAINT_WINDOW_TRANSLUCENT_MASK)
+	if (mask & PAINT_WINDOW_BLEND_MASK)
 	    glDisable (GL_BLEND);
     }
     else
     {
-	enableTexture (w->screen, texture, filter);
+	enableTexture (s, texture, filter);
 
-	if (mask & PAINT_WINDOW_TRANSLUCENT_MASK)
+	if (mask & PAINT_WINDOW_BLEND_MASK)
 	{
 	    glEnable (GL_BLEND);
 	    if (attrib->opacity != OPAQUE || attrib->brightness != BRIGHT)
@@ -764,48 +891,75 @@ drawWindowTexture (CompWindow		   *w,
 
 		color = (attrib->opacity * attrib->brightness) >> 16;
 
-		screenTexEnvMode (w->screen, GL_MODULATE);
+		screenTexEnvMode (s, GL_MODULATE);
 		glColor4us (color, color, color, attrib->opacity);
 
-		(*w->screen->drawWindowGeometry) (w);
+		(*w->drawWindowGeometry) (w);
 
 		glColor4usv (defaultColor);
-		screenTexEnvMode (w->screen, GL_REPLACE);
+		screenTexEnvMode (s, GL_REPLACE);
 	    }
 	    else
 	    {
-		(*w->screen->drawWindowGeometry) (w);
+		(*w->drawWindowGeometry) (w);
 	    }
 
 	    glDisable (GL_BLEND);
 	}
 	else if (attrib->brightness != BRIGHT)
 	{
-	    screenTexEnvMode (w->screen, GL_MODULATE);
+	    screenTexEnvMode (s, GL_MODULATE);
 	    glColor4us (attrib->brightness, attrib->brightness,
 			attrib->brightness, BRIGHT);
 
-	    (*w->screen->drawWindowGeometry) (w);
+	    (*w->drawWindowGeometry) (w);
 
 	    glColor4usv (defaultColor);
-	    screenTexEnvMode (w->screen, GL_REPLACE);
+	    screenTexEnvMode (s, GL_REPLACE);
 	}
 	else
 	{
-	    (*w->screen->drawWindowGeometry) (w);
+	    (*w->drawWindowGeometry) (w);
 	}
 
 	disableTexture (w->screen, texture);
     }
+}
 
-    glPopMatrix ();
+void
+drawWindowTexture (CompWindow		*w,
+		   CompTexture		*texture,
+		   const FragmentAttrib	*attrib,
+		   unsigned int		mask)
+{
+    int filter;
+
+    if (mask & (PAINT_WINDOW_TRANSFORMED_MASK |
+		PAINT_WINDOW_ON_TRANSFORMED_SCREEN_MASK))
+	filter = w->screen->filter[SCREEN_TRANS_FILTER];
+    else
+	filter = w->screen->filter[NOTHING_TRANS_FILTER];
+
+    if (!attrib->nFunction || !enableFragmentProgramAndDrawGeometry (w,
+								     texture,
+								     attrib,
+								     filter,
+								     mask))
+    {
+	enableFragmentOperationsAndDrawGeometry (w,
+						 texture,
+						 attrib,
+						 filter,
+						 mask);
+    }
 }
 
 Bool
-drawWindow (CompWindow		    *w,
-	    const WindowPaintAttrib *attrib,
-	    Region		    region,
-	    unsigned int	    mask)
+drawWindow (CompWindow		 *w,
+	    const CompTransform  *transform,
+	    const FragmentAttrib *fragment,
+	    Region		 region,
+	    unsigned int	 mask)
 {
     if (mask & PAINT_WINDOW_TRANSFORMED_MASK)
 	region = &infiniteRegion;
@@ -813,20 +967,12 @@ drawWindow (CompWindow		    *w,
     if (!region->numRects)
 	return TRUE;
 
-    if (mask & PAINT_WINDOW_SOLID_MASK)
+    if (mask & PAINT_WINDOW_CLIP_OPAQUE_MASK)
     {
 	if (w->attrib.map_state != IsViewable)
 	    return FALSE;
 
-	if (w->alpha)
-	    return FALSE;
-
-	if (attrib->opacity != OPAQUE)
-	    return FALSE;
-    }
-    else if (mask & PAINT_WINDOW_TRANSLUCENT_MASK)
-    {
-	if (!w->alpha && attrib->opacity == OPAQUE)
+	if (mask & PAINT_WINDOW_TRANSLUCENT_MASK)
 	    return FALSE;
     }
     else
@@ -834,23 +980,23 @@ drawWindow (CompWindow		    *w,
 	if (w->attrib.map_state != IsViewable)
 	    return TRUE;
 
-	if (w->alpha || attrib->opacity != OPAQUE)
-	    mask |= PAINT_WINDOW_TRANSLUCENT_MASK;
-	else
-	    mask |= PAINT_WINDOW_SOLID_MASK;
+	if (mask & PAINT_WINDOW_CLIP_TRANSLUCENT_MASK)
+	{
+	    if (!(mask & PAINT_WINDOW_TRANSLUCENT_MASK))
+		return FALSE;
+	}
     }
 
-    if (!w->texture->pixmap)
-    {
-	bindWindow (w);
-	if (!w->mapNum)
-	    return FALSE;
-    }
+    if (!w->texture->pixmap && !bindWindow (w))
+	return FALSE;
 
-    w->vCount = 0;
+    if (mask & PAINT_WINDOW_TRANSLUCENT_MASK)
+	mask |= PAINT_WINDOW_BLEND_MASK;
+
+    w->vCount = w->indexCount = 0;
     (*w->screen->addWindowGeometry) (w, &w->matrix, 1, w->region, region);
     if (w->vCount)
-	(*w->screen->drawWindowTexture) (w, w->texture, attrib, mask);
+	(*w->screen->drawWindowTexture) (w, w->texture, fragment, mask);
 
     return TRUE;
 }
@@ -858,20 +1004,40 @@ drawWindow (CompWindow		    *w,
 Bool
 paintWindow (CompWindow		     *w,
 	     const WindowPaintAttrib *attrib,
+	     const CompTransform     *transform,
 	     Region		     region,
 	     unsigned int	     mask)
 {
+    FragmentAttrib fragment;
+    Bool	   status;
+
     w->lastPaint = *attrib;
 
     if (mask & PAINT_WINDOW_NO_CORE_INSTANCE_MASK)
     {
-	if (mask & PAINT_WINDOW_SOLID_MASK)
+	if (mask & PAINT_WINDOW_CLIP_OPAQUE_MASK)
 	    return FALSE;
 
 	return TRUE;
     }
 
-    return (*w->screen->drawWindow) (w, attrib, region, mask);
+    initFragmentAttrib (&fragment, attrib);
+
+    if (w->alpha || fragment.opacity != OPAQUE)
+	mask |= PAINT_WINDOW_TRANSLUCENT_MASK;
+
+    if (mask & PAINT_WINDOW_TRANSFORMED_MASK)
+    {
+	glPushMatrix ();
+	glLoadMatrixf (transform->m);
+    }
+
+    status = (*w->screen->drawWindow) (w, transform, &fragment, region, mask);
+
+    if (mask & PAINT_WINDOW_TRANSFORMED_MASK)
+	glPopMatrix ();
+
+    return status;
 }
 
 void
@@ -897,8 +1063,7 @@ paintBackground (CompScreen   *s,
 
 	s->backgroundLoaded = FALSE;
 
-	if (!(mask & PAINT_BACKGROUND_WITH_STENCIL_MASK))
-	    return;
+	return;
     }
     else
     {
@@ -943,13 +1108,6 @@ paintBackground (CompScreen   *s,
 	pBox++;
     }
 
-    if (mask & PAINT_BACKGROUND_WITH_STENCIL_MASK)
-    {
-	glEnable (GL_STENCIL_TEST);
-	glStencilFunc (GL_ALWAYS, s->stencilRef, ~0);
-	glStencilOp (GL_REPLACE, GL_REPLACE, GL_REPLACE);
-    }
-
     glTexCoordPointer (2, GL_FLOAT, sizeof (GLfloat) * 4, data);
     glVertexPointer (2, GL_FLOAT, sizeof (GLfloat) * 4, data + 2);
 
@@ -978,12 +1136,5 @@ paintBackground (CompScreen   *s,
 	}
     }
 
-    if (mask & PAINT_BACKGROUND_WITH_STENCIL_MASK)
-    {
-	glStencilFunc (GL_EQUAL, s->stencilRef, ~0);
-	glStencilOp (GL_KEEP, GL_KEEP, GL_KEEP);
-	glDisable (GL_STENCIL_TEST);
-    }
-
     free (data);
 }
