@@ -75,6 +75,8 @@
 
 #define OUTPUTS_DEFAULT "640x480+0+0"
 
+#define FOCUS_PREVENTION_DEFAULT TRUE
+
 #define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))
 
 static int
@@ -460,6 +462,7 @@ setScreenOption (CompScreen      *screen,
     case COMP_SCREEN_OPTION_LIGHTING:
     case COMP_SCREEN_OPTION_UNREDIRECT_FS:
     case COMP_SCREEN_OPTION_SYNC_TO_VBLANK:
+    case COMP_SCREEN_OPTION_FOCUS_PREVENTION:
 	if (compSetBoolOption (o, value))
 	    return TRUE;
 	break;
@@ -680,6 +683,13 @@ compScreenInitOptions (CompScreen *screen)
     o->value.list.value->s = strdup (OUTPUTS_DEFAULT);
     o->rest.s.string       = NULL;
     o->rest.s.nString      = 0;
+
+    o = &screen->opt[COMP_SCREEN_OPTION_FOCUS_PREVENTION];
+    o->name      = "focus_prevention";
+    o->shortDesc = N_("Focus prevention");
+    o->longDesc  = N_("Enable focus prevention");
+    o->type      = CompOptionTypeBool;
+    o->value.b   = FOCUS_PREVENTION_DEFAULT;
 }
 
 static void
@@ -835,47 +845,47 @@ updateScreenEdges (CompScreen *s)
 }
 
 static void
-frustum (GLfloat left,
+frustum (GLfloat *m,
+	 GLfloat left,
 	 GLfloat right,
 	 GLfloat bottom,
 	 GLfloat top,
 	 GLfloat nearval,
 	 GLfloat farval)
 {
-   GLfloat x, y, a, b, c, d;
-   GLfloat m[16];
+    GLfloat x, y, a, b, c, d;
 
-   x = (2.0 * nearval) / (right - left);
-   y = (2.0 * nearval) / (top - bottom);
-   a = (right + left) / (right - left);
-   b = (top + bottom) / (top - bottom);
-   c = -(farval + nearval) / ( farval - nearval);
-   d = -(2.0 * farval * nearval) / (farval - nearval);
+    x = (2.0 * nearval) / (right - left);
+    y = (2.0 * nearval) / (top - bottom);
+    a = (right + left) / (right - left);
+    b = (top + bottom) / (top - bottom);
+    c = -(farval + nearval) / ( farval - nearval);
+    d = -(2.0 * farval * nearval) / (farval - nearval);
 
 #define M(row,col)  m[col*4+row]
-   M(0,0) = x;     M(0,1) = 0.0F;  M(0,2) = a;      M(0,3) = 0.0F;
-   M(1,0) = 0.0F;  M(1,1) = y;     M(1,2) = b;      M(1,3) = 0.0F;
-   M(2,0) = 0.0F;  M(2,1) = 0.0F;  M(2,2) = c;      M(2,3) = d;
-   M(3,0) = 0.0F;  M(3,1) = 0.0F;  M(3,2) = -1.0F;  M(3,3) = 0.0F;
+    M(0,0) = x;     M(0,1) = 0.0f;  M(0,2) = a;      M(0,3) = 0.0f;
+    M(1,0) = 0.0f;  M(1,1) = y;     M(1,2) = b;      M(1,3) = 0.0f;
+    M(2,0) = 0.0f;  M(2,1) = 0.0f;  M(2,2) = c;      M(2,3) = d;
+    M(3,0) = 0.0f;  M(3,1) = 0.0f;  M(3,2) = -1.0f;  M(3,3) = 0.0f;
 #undef M
 
-   glMultMatrixf (m);
 }
 
 static void
-perspective (GLfloat fovy,
+perspective (GLfloat *m,
+	     GLfloat fovy,
 	     GLfloat aspect,
 	     GLfloat zNear,
 	     GLfloat zFar)
 {
-   GLfloat xmin, xmax, ymin, ymax;
+    GLfloat xmin, xmax, ymin, ymax;
 
-   ymax = zNear * tan (fovy * M_PI / 360.0);
-   ymin = -ymax;
-   xmin = ymin * aspect;
-   xmax = ymax * aspect;
+    ymax = zNear * tan (fovy * M_PI / 360.0);
+    ymin = -ymax;
+    xmin = ymin * aspect;
+    xmax = ymax * aspect;
 
-   frustum (xmin, xmax, ymin, ymax, zNear, zFar);
+    frustum (m, xmin, xmax, ymin, ymax, zNear, zFar);
 }
 
 void
@@ -909,9 +919,11 @@ reshape (CompScreen *s,
 
     s->rasterX = s->rasterY = 0;
 
+    perspective (s->projection, 60.0f, 1.0f, 0.1f, 100.0f);
+
     glMatrixMode (GL_PROJECTION);
     glLoadIdentity ();
-    perspective (60.0f, 1.0f, 0.1f, 100.0f);
+    glMultMatrixf (s->projection);
     glMatrixMode (GL_MODELVIEW);
 
     s->region.rects = &s->region.extents;
@@ -1385,7 +1397,6 @@ addScreen (CompDisplay *display,
     unsigned int	 nchildren;
     int			 defaultDepth, nvisinfo, nElements, value, i;
     const char		 *glxExtensions, *glExtensions;
-    GLint	         stencilBits;
     XSetWindowAttributes attrib;
     GLfloat		 globalAmbient[]  = { 0.1f, 0.1f,  0.1f, 0.1f };
     GLfloat		 ambientLight[]   = { 0.0f, 0.0f,  0.0f, 0.0f };
@@ -1487,8 +1498,6 @@ addScreen (CompDisplay *display,
     s->windows = 0;
     s->reverseWindows = 0;
 
-    s->stencilRef = 0x1;
-
     s->opacityStep = OPACITY_STEP_DEFAULT;
 
     s->nextRedraw  = 0;
@@ -1499,6 +1508,13 @@ addScreen (CompDisplay *display,
 
     s->pendingCommands = TRUE;
 
+    s->lastFunctionId = 0;
+
+    s->fragmentFunctions = NULL;
+    s->fragmentPrograms = NULL;
+
+    memset (s->saturateFunction, 0, sizeof (s->saturateFunction));
+
     s->showingDesktopMask = 0;
 
     s->overlayWindowCount = 0;
@@ -1526,7 +1542,6 @@ addScreen (CompDisplay *display,
     s->drawWindow		 = drawWindow;
     s->addWindowGeometry	 = addWindowGeometry;
     s->drawWindowTexture	 = drawWindowTexture;
-    s->drawWindowGeometry	 = drawWindowGeometry;
     s->damageWindowRect		 = damageWindowRect;
     s->getOutputExtentsForWindow = getOutputExtentsForWindow;
     s->focusWindow		 = focusWindow;
@@ -1792,13 +1807,16 @@ addScreen (CompDisplay *display,
 	    getProcAddress (s, "glBindProgramARB");
 	s->programString = (GLProgramStringProc)
 	    getProcAddress (s, "glProgramStringARB");
-	s->programLocalParameter4f = (GLProgramLocalParameter4fProc)
+	s->programEnvParameter4f = (GLProgramParameter4fProc)
+	    getProcAddress (s, "glProgramEnvParameter4fARB");
+	s->programLocalParameter4f = (GLProgramParameter4fProc)
 	    getProcAddress (s, "glProgramLocalParameter4fARB");
 
-	if (s->genPrograms    &&
-	    s->deletePrograms &&
-	    s->bindProgram    &&
-	    s->programString  &&
+	if (s->genPrograms	     &&
+	    s->deletePrograms	     &&
+	    s->bindProgram	     &&
+	    s->programString	     &&
+	    s->programEnvParameter4f &&
 	    s->programLocalParameter4f)
 	    s->fragmentProgram = 1;
     }
@@ -1975,14 +1993,6 @@ addScreen (CompDisplay *display,
 
     s->desktopWindowCount = 0;
 
-    glGetIntegerv (GL_STENCIL_BITS, &stencilBits);
-    if (!stencilBits)
-    {
-	fprintf (stderr, "%s: No stencil buffer. Clipping of transformed "
-		 "windows is not going to be correct when screen is "
-		 "transformed.\n", programName);
-    }
-
     glClearColor (0.0, 0.0, 0.0, 1.0);
     glBlendFunc (GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
     glEnable (GL_CULL_FACE);
@@ -2007,6 +2017,7 @@ addScreen (CompDisplay *display,
 
     detectRefreshRateOfScreen (s);
     detectOutputDevices (s);
+    updateOutputDevices (s);
 
     glLightModelfv (GL_LIGHT_MODEL_AMBIENT, globalAmbient);
 
@@ -3125,9 +3136,9 @@ moveScreenViewport (CompScreen *s,
 	else
 	{
 	    m = w->attrib.x + tx;
-	    if (m - w->output.left < s->width - vWidth)
+	    if (m - w->input.left < s->width - vWidth)
 		wx = tx + vWidth;
-	    else if (m + w->width + w->output.right > vWidth)
+	    else if (m + w->width + w->input.right > vWidth)
 		wx = tx - vWidth;
 	    else
 		wx = tx;
@@ -3144,9 +3155,9 @@ moveScreenViewport (CompScreen *s,
 	else
 	{
 	    m = w->attrib.y + ty;
-	    if (m - w->output.top < s->height - vHeight)
+	    if (m - w->input.top < s->height - vHeight)
 		wy = ty + vHeight;
-	    else if (m + w->height + w->output.bottom > vHeight)
+	    else if (m + w->height + w->input.bottom > vHeight)
 		wy = ty - vHeight;
 	    else
 		wy = ty;
@@ -3303,7 +3314,11 @@ applyStartupProperties (CompScreen *screen,
 	window->initialViewportX = s->viewportX;
 	window->initialViewportY = s->viewportY;
 
-	window->desktop = sn_startup_sequence_get_workspace (s->sequence);
+	window->desktop	= sn_startup_sequence_get_workspace (s->sequence);
+
+	window->initialTimestamp    =
+	    sn_startup_sequence_get_timestamp (s->sequence);
+	window->initialTimestampSet = TRUE;
     }
 }
 
@@ -3664,6 +3679,89 @@ clearScreenOutput (CompScreen	*s,
     }
 }
 
+/* Returns default viewport for some window geometry. If the window spans
+   more than one viewport the most appropriate viewport is returned. How the
+   most appropriate viewport is computed can be made optional if necessary. It
+   is currently computed as the viewport where the center of the window is
+   located, except for when the window is visible in the current viewport as
+   the current viewport is then always returned. */
+void
+viewportForGeometry (CompScreen *s,
+		     int	x,
+		     int	y,
+		     int	width,
+		     int	height,
+		     int	borderWidth,
+		     int	*viewportX,
+		     int	*viewportY)
+{
+    int	centerX;
+    int	centerY;
+
+    width  += borderWidth * 2;
+    height += borderWidth * 2;
+
+    if ((x < s->width  && x + width  > 0) &&
+	(y < s->height && y + height > 0))
+    {
+	if (viewportX)
+	    *viewportX = s->x;
+
+	if (viewportY)
+	    *viewportY = s->y;
+
+	return;
+    }
+
+    if (viewportX)
+    {
+	centerX = x + (width >> 1);
+	if (centerX < 0)
+	    *viewportX = s->x + ((centerX / s->width) - 1) % s->hsize;
+	else
+	    *viewportX = s->x + (centerX / s->width) % s->hsize;
+    }
+
+    if (viewportY)
+    {
+	centerY = y + (height >> 1);
+	if (centerY < 0)
+	    *viewportY = s->y + ((centerY / s->height) - 1) % s->vsize;
+	else
+	    *viewportY = s->y + (centerY / s->height) % s->vsize;
+    }
+}
+
+int
+outputDeviceForGeometry (CompScreen *s,
+			 int	    x,
+			 int	    y,
+			 int	    width,
+			 int	    height,
+			 int	    borderWidth)
+{
+    int output = s->currentOutputDev;
+    int x1, y1, x2, y2;
+
+    width  += borderWidth * 2;
+    height += borderWidth * 2;
+
+    x1 = s->outputDev[output].region.extents.x1;
+    y1 = s->outputDev[output].region.extents.y1;
+    x2 = s->outputDev[output].region.extents.x2;
+    y2 = s->outputDev[output].region.extents.y2;
+
+    if (x1 > x + width  ||
+	y1 > y + height ||
+	x2 < x		||
+	y2 < y)
+    {
+	output = outputDeviceForPoint (s, x + width  / 2, y + height / 2);
+    }
+
+    return output;
+}
+
 Bool
 updateDefaultIcon (CompScreen *screen)
 {
